/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#ifndef		Q_EVOLUTIONARY_ALGORITHM_H
#define		Q_EVOLUTIONARY_ALGORITHM_H

#include "../../qtools/qtools.h"
#include "../operators/include.h"
#include "../encodings/include.h"
#include "../include/qec.h"
#include "../include/qrunsummary.h"

#include <fstream>

namespace ECWorkshop {

///////////////////////////////////////////////////////////
//
//		QEvolutionaryAlgorithm class
//
// Abstract class to define general properties of EAs' run
//
///////////////////////////////////////////////////////////

class QEvolutionaryAlgorithm {
protected:
	QEAParameters _params;
	QRunSummary _summary;
	QEAFactory* _factory;
	QPopulation* _popul;
	QPopulation* _selPopul;
	QPopulation* _children;
	QInitializer* _init;
	QNichingOperator* _niching;
	QSelection* _selection;
	QPopulationSizer* _populationSizer;
	QCrossoverOperator* _xover;
	QMutationOperator* _mutation;

	ofstream _fout;

	QEvolutionaryAlgorithm (const QEvolutionaryAlgorithm& argEA);
	virtual QEvolutionaryAlgorithm& operator= (const QEvolutionaryAlgorithm& argEA);

public:
	QEvolutionaryAlgorithm (void);
	virtual ~QEvolutionaryAlgorithm(void);

	// int setParameters (const QGAParameters& argParams)
	//	Sets parameters of the EA and makes necessary initializations
	//	Returns Q_OK on success
	virtual int setParameters (const QEAParameters& argParams);

	virtual QRunSummary& getSummary ();

	QPopulation* getPopulation () const;

	// int initializeComponents ()
	//	Makes initialization of EA components without changes
	//	in parameters.
	//	Returns Q_OK on success
	virtual int initializeComponents ();

	// int evaluatePopulation
	// Performs evaluation of population using
	//  [_params.problem] instance of [QProblem] object
	virtual int evaluatePopulation ();
	// int evaluatePopulation
	// Performs evaluation of population using
	//  [_params.problem] instance of [QProblem] object
	//	without writing evaluation results statistics
	virtual int evaluatePopulationNoStats ();

	// int selectIndividuals
	// Performs selections
	virtual int selectParentIndividuals ();

	// int selectElite
	// Selects [_params::eliteCount] elite parent individuals
	virtual int selectElite ();

	// int crossPopulation
	// Makes crossover operator to work
	virtual int crossPopulation ();

	// int mutatePopulation ()
	// Makes mutation operator to work
	virtual int mutatePopulation ();
	QMutationOperator* getMutationOperator () const;

	QNichingOperator* getNichingOperator () const;

	QEAFactory* getFactory () const;

	// int nextGeneration ()
	// Proceeds to the next generation via definition of _popul
	//  using _children and, probably, _popul and _selPopul
	//  population containers
	virtual int nextGeneration ();

	// int printName (ostream& out)
	//	Prints algorithm name into the specified output stream
	virtual int printName (ostream& out) = 0;

	// int printResults (__VECTOR_RUN_SUMMARY& argS, ostream& out)
	//	Prints algorithm specific results into the passed output stream
	virtual int printResults (__VECTOR_RUN_SUMMARY& argS, ostream& out) = 0;

	// int finishWork ()
	// Makes some finalization actions (deallocation of memory etc)
	//	should be called after all runs is finished and data is processed
	//	If not called than memory leaks are possible
	virtual int finishWork ();

	// int run ()
	//	Performs the main routine of evolutionary algorithm/model
	//	Returns Q_OK on success
	virtual int run () = 0;

	// QEvolutionaryAlgorithm* clone ()
	//	Makes copy of the existing object
	virtual QEvolutionaryAlgorithm* clone () = 0;
};

typedef	vector<QEvolutionaryAlgorithm*>	__VECTOR_EA_PTR;

};	// namespace

#endif
