/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qsarn.h"
#include "../problems/qannproblem.h"

using namespace ECWorkshop;

QSARN::QSARN (const QSARN& argEA) : QEvolutionaryAlgorithm(argEA) {
}

QSARN& QSARN::operator= (const QSARN& argEA) {
	// no assignment is made
	return *this;
}

QSARN::QSARN (void) {
	_fout.open ("sarn.log");
}

QSARN::~QSARN (void) {
	_fout.close ();
}

// int setParameters (const QGAParameters& argParams)
//	Sets parameters of the EA and makes necessary initializations
//	Returns Q_OK on success
int QSARN::setParameters (const QEAParameters& argParams) {
	this->QEvolutionaryAlgorithm::setParameters (argParams);

	return Q_OK;
}

// int run ()
//	Performs the main routine of evolutionary algorithm/model
//	Returns Q_OK on success
int QSARN::run () {
	if (_params.generationsNumber > 0) {
		int i;

		_summary.reset ();
		_summary.startClock = clock ();

		initializeComponents ();

		_fout<<"\nStarting a new run:\n"
			<<"-------------------------------\n";
		_popul->printPopulation (_fout);
		_fout.flush ();

		// zero [xRate] to disable crossover operator
		_params.xRate = 0.0;
		for (i=0; i<_params.generationsNumber; i++) {

			_summary.popSizes[ _summary.generation ] = _popul->getSize();
			//cout<<"evaluatePopulation\n";
			//cout.flush ();
			if (evaluatePopulation () == Q_OK) {
				//cout<<"selectParentIndividuals\n";
				//cout.flush ();
				selectParentIndividuals ();
	//			cout<<"selectElite\n";
	//			cout.flush ();
				selectElite ();
				//cout<<"crossPopulation\n";
				//cout.flush ();
				crossPopulation ();
				//cout<<"mutatePopulation\n";
				//cout.flush ();
				mutatePopulation ();
				//cout<<"nextGeneration\n";
				//cout.flush ();
				nextGeneration ();

				_summary.generation++;

				if (! (_summary.generation % 10)) {
					printf(".");
				}
			} else {	// first hit event occured with the [QEAParameters::untilFirstHit] flag set
				break;
			}
		}
		printf ("\nFinishing the run...");

		_summary.finishClock = clock ();
		_summary.time = (double) (_summary.finishClock - _summary.startClock) / CLOCKS_PER_SEC;

		printf("done\n");

		evaluatePopulationNoStats ();
		_popul->printPopulation (_fout);

		// test found solution or current best individual
		if (_summary.firstHitSolution != NULL) {
			((QANNProblem*)_params.problem)->testIndividual (_summary.firstHitSolution);
		} else {
			((QANNProblem*)_params.problem)->testIndividual (_summary.currentBest);
		}

		_fout<<"Run duration: "<<_summary.time<<" seconds\n";

		return Q_OK;
	} else {
		cerr<<"\n\n[QGeneticAlgorithm::run] error:\n"
			<<"\tGenerations number is undefined! (_params.generationsNumber = "<<_params.generationsNumber<<")\n\n";
		return Q_FAILURE;
	}
	return Q_OK;
}

// QEvolutionaryAlgorithm* clone ()
//	Makes copy of the existing object
QEvolutionaryAlgorithm* QSARN::clone () {
	return new QSARN (*this);
}

// int crossPopulation
// Overloaded method
int QSARN::crossPopulation () {
	int size = _selPopul->getSize();
	_children->assign (_selPopul);
	while (!_populationSizer->populationReady()) {
		_children->addIndividual( _selPopul->getIndividual( (int)(QRandN() * (double)size) ) );
	}
	_populationSizer->operate ();

	return Q_OK;
}

// int printResults (__VECTOR_RUN_SUMMARY& argS, ostream& out)
//	Prints algorithm specific results into the passed output stream
int QSARN::printResults (__VECTOR_RUN_SUMMARY& argS, ostream& out) {
	int sols;	// solutions number
	int i, j, size = (int)argS.size(), temp, size2;
	double mean, dev, meanC, devC;
	__VECTOR_INT tempV, tempD;
	__VECTOR_DOUBLE mutS, mutC;

	// calculate number of solutions
	// and average number of nodes pur population of solutions
	// and population of candidates ([meanC] and [devC])
	sols = 0;
	mean = dev = 0.0;
	meanC = devC = 0.0;
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitFECount != 0) {
			// calculate stats for solutions
			++sols;
			temp = argS[i].firstHitSolution->getSize();
			mean += temp;
			dev += temp*temp;
		} else {
			// calculate stats for candidates
			temp = argS[i].currentBest->getSize();
			meanC += temp;
			devC += temp*temp;
		}
	}
	mean /= (double)sols;
	dev = dev/(double)sols - mean*mean;
	dev *= (double)sols/(sols-1);
	if (sols < size) {
		meanC /= (double)(size-sols);
		devC = devC/(double)(size-sols) - meanC*meanC;
		devC *= (double)(size-sols)/((size-sols)-1);
	}

	// print average number of nodes
	out<<"> Solutions summary ("<<sols<<"):\n"
		<<"Average number of nodes:\t"<<mean<<"\n"
		<<"Deviation of the number of nodes:\t"<<dev<<"\n";
	if (sols < size) {
		out<<"> Candidate-solutions summary ("<<(size-sols)<<"):\n"
			<<"Average number of nodes:\t"<<meanC<<"\n"
			<<"Deviation of the number of nodes:\t"<<devC<<"\n";
	}

	// calculate data about distribution of nodes degrees
	mean = dev = 0.0;
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitTime != 0) {
			((QRNIndividual*)(argS[i].firstHitSolution))->calculateDegrees (tempV);

			size2 = (int)tempV.size ();
			for (j=0; j<size2; ++j) {
				while ((int)tempD.size() <= tempV[j]) {
					tempD.push_back (0);
				}
				tempD[ tempV[j] ]++;
			}

			// calculate average and deviation for nodes degrees
			qstats (tempV, meanC, devC);
			mean += meanC;
			dev += devC;
		}
	}
	if (sols > 0) {
		mean /= (double)sols;
		dev /= (double)sols;
		out<<"> Distribution of nodes degrees over the population of solutions:\t";
		size2 = (int)tempD.size ();
		for (i=0; i<size2; ++i) {
			out<<" "<<tempD[i];
		}
		out<<"\n";
		out<<"\tMean nodes degree:\t"<<mean<<"\n"
			<<"\tAveraged deviation of the nodes degree:\t"<<dev<<"\n";
	}

	// calculate averaged vector of mutation parameters
	mutS.assign (MUTATION_PARAMETERS, 0.0);
	mutC.assign (MUTATION_PARAMETERS, 0.0);
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitTime != 0) {
			// calculate history for solutions
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				mutS[j] += ((QRNIndividual*)(argS[i].firstHitSolution))->getMutationParam (j);
			}
		} else {
			// calculate history for candidates
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				mutC[j] += ((QRNIndividual*)(argS[i].currentBest))->getMutationParam (j);
			}
		}
	}
	if (sols > 0) {
		out<<"> Averaged vector of mutation parameters for the solutions:\t";
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			out<<mutS[i]/(double)sols<<" ";
		}
		out<<"\n";
	}
	if (sols < size) {
		out<<"> Averaged vector of mutation parameters for the candidates:\t";
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			out<<mutC[i]/(double)(size-sols)<<" ";
		}
		out<<"\n";
	}

	// calculate cumulative history events
	tempV.assign (MUTATION_PARAMETERS, 0);
	tempD.assign (MUTATION_PARAMETERS, 0);
	for (i=0; i<size; ++i) {
		if (argS[i].firstHitTime != 0) {
			// calculate history for solutions
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				tempV[j] += ((QRNIndividual*)(argS[i].firstHitSolution))->readHistory (j);
			}
		} else {
			// calculate history for candidates
			for (j=0; j<MUTATION_PARAMETERS; ++j) {
				tempD[j] += ((QRNIndividual*)(argS[i].currentBest))->readHistory (j);
			}
		}
	}
	out<<"> Cumulative history events counters for the solutions:\t";
	for (i=0; i<MUTATION_PARAMETERS; ++i) {
		out<<tempV[i]<<" ";
	}
	out<<"\n";
	if (sols < size) {
		out<<"> Cumulative history events counters for the candidates:\t";
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			out<<tempD[i]<<" ";
		}
		out<<"\n";
	}

	return Q_OK;
}

// int printName (ostream& out)
//	Prints algorithm name into the specified output stream
int QSARN::printName (ostream& out) {
	out<<"Self-adaptive regulatory network\n";
	return Q_OK;
}
