/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/


#include "qintegerindividual.h"

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//		QIntegerIndividual class
//
///////////////////////////////////////////////////////////

QIntegerIndividual::QIntegerIndividual(void) {
}

QIntegerIndividual::QIntegerIndividual (const QIntegerIndividual& argInd)
: QIndividual(argInd) {
	_genes.assign (argInd._genes.begin(), argInd._genes.end());
}

QIntegerIndividual::~QIntegerIndividual(void) {
	clearIndividual ();
}

QIntegerIndividual& QIntegerIndividual::operator = (const QIntegerIndividual& argInd) {
	this->QIndividual::operator = (argInd);
	_genes.assign (argInd._genes.begin(), argInd._genes.end());
	return (*this);
}

int QIntegerIndividual::assign (const QIndividual* argInd) {
	this->operator= ((QIntegerIndividual&)(*argInd));
	return Q_OK;
}

int QIntegerIndividual::getSize () const {
	return (int)_genes.size ();
}

int QIntegerIndividual::addGene (const QGene* argGene) {
	_genes.push_back ((QIntegerGene&)(*argGene));
	return (int)_genes.size ();
}

int QIntegerIndividual::deleteGene (const int argIndex) {
	if (argIndex < (int)_genes.size ()) {

		_genes.erase (_genes.begin() + argIndex);
		return (int)_genes.size ();

	} else {

		cerr<<"\n\n[QIntegerIndividual::deleteGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);

	}
}

QGene* QIntegerIndividual::getGene (const int argIndex) {
	if (argIndex < (int)_genes.size ()) {
		return &_genes[ argIndex ];
	} else {
		cerr<<"\n\n[QIntegerIndividual::getGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

const QGene* QIntegerIndividual::getGene (const int argIndex) const {
	if (argIndex < (int)_genes.size ()) {
		return &_genes[ argIndex ];
	} else {
		cerr<<"\n\n[QIntegerIndividual::getGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

int QIntegerIndividual::setGene (const int argIndex, const QGene* argGene) {
	if (argIndex < (int)_genes.size ()) {
		_genes[ argIndex ].assign (argGene);
		return Q_OK;
	} else {
		cerr<<"\n\n[QIntegerIndividual::setGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

int QIntegerIndividual::clearIndividual () {
	_genes.clear ();
	return Q_OK;
}

bool QIntegerIndividual::equalsTo (const QIndividual* argInd) const {
	if (getEncodingType() == argInd->getEncodingType()) {
		const QIntegerIndividual* tempInd = (QIntegerIndividual*)argInd;
		if (_genes.size() == tempInd->_genes.size()) {
			__VECTOR_INTEGER_GENE::const_iterator iter = _genes.begin (), end = _genes.end();
			__VECTOR_INTEGER_GENE::const_iterator argIter = tempInd->_genes.begin ();

			for (; iter != end; iter++, argIter++) {
				if ((*iter).equalsTo (&(*argIter)) == false) {
					return false;
				}
			}
			return true;
		}
	}

	return false;
}

int QIntegerIndividual::getGeneSize () const {
	return _params->geneSize;
}

QIndividual* QIntegerIndividual::clone () const {
	return new QIntegerIndividual ((QIntegerIndividual&)(*this));
}

QEncodingType QIntegerIndividual::getEncodingType () const {
	return QET_INTEGER;
}

int QIntegerIndividual::getGeneValue (const int argIndex) const {
	return _genes[ argIndex ].getValue ();
}

int QIntegerIndividual::setGeneValue (const int argIndex, const int argValue) {
	if (argIndex < (int)_genes.size () && argIndex >= 0) {
		_genes[ argIndex ].setValue (argValue);
		return Q_OK;
	} else {
		cerr<<"\n\nQIntegerIndividual::setGeneValue error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

int QIntegerIndividual::createIndividual (const int argGenesNumber) {
	_genes.resize (argGenesNumber);
	return Q_OK;
}

int QIntegerIndividual::print2Stream (ostream& out) const {
	__VECTOR_INTEGER_GENE::const_iterator iter = _genes.begin (), end = _genes.end();
	const QIntegerGene *tempGene = NULL;

	out<<"("<<_fitness.getValue()<<")\t";
	out<<hex;
	for (; iter != end; iter++) {
		tempGene = &(*iter);
		out<<tempGene->getValue()<<" ";
	}
	out<<dec;

	return Q_OK;
}

int QIntegerIndividual::toVectorDouble (__VECTOR_DOUBLE& target) {
	int size = (int)_genes.size();
	target.resize (size);
	for (int i=0; i<size; i++) {
		target[i] = _genes[i].getValue()*_params->genePrecision + _params->geneOffset;
	}
	return Q_OK;
}

int QIntegerIndividual::toVectorInt (__VECTOR_INT& target) {
	if (_params->genePrecision == 1 && ((int)_params->geneOffset == _params->geneOffset)) {	// if genes encode integer parameters
		int size = (int)_genes.size();
		target.resize (size);
		for (int i=0; i<size; i++) {
			target[i] = (int)(_genes[i].getValue() + _params->geneOffset);
		}
		return Q_OK;
	} else {	// conversion involves loss of precision
		return Q_FAILURE;
	}
}

int QIntegerIndividual::toVectorDouble (__VECTOR_DOUBLE& target) const {
	int size = (int)_genes.size();
	target.resize (size);
	for (int i=0; i<size; i++) {
		target[i] = _genes[i].getValue()*_params->genePrecision + _params->geneOffset;
	}
	return Q_OK;
}

int QIntegerIndividual::toVectorInt (__VECTOR_INT& target) const {
	if (_params->genePrecision == 1 && ((int)_params->geneOffset == _params->geneOffset)) {	// if genes encode integer parameters
		int size = (int)_genes.size();
		target.resize (size);
		for (int i=0; i<size; i++) {
			target[i] = (int)(_genes[i].getValue() + _params->geneOffset);
		}
		return Q_OK;
	} else {	// conversion involves loss of precision
		return Q_FAILURE;
	}
}
