/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qpopulation.h"

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//		QPopulation class
//
///////////////////////////////////////////////////////////

QPopulation::QPopulation (void) {
}

QPopulation::QPopulation (const QPopulation& argPopul) {
	__VECTOR_INDIVIDUAL_PTR::const_iterator iter = argPopul._inds.begin (), end = argPopul._inds.end();
	_inds.reserve (argPopul.getSize());
	for (; iter != end; iter++) {
		_inds.push_back ((*iter)->clone ());
	}
}

QPopulation::~QPopulation (void) {
	clearPopulation ();
}

QPopulation& QPopulation::operator = (const QPopulation& argPopul) {
	if ((int)_inds.size () == argPopul.getSize()) {
		__VECTOR_INDIVIDUAL_PTR::iterator iter = _inds.begin (), end = _inds.end();
		__VECTOR_INDIVIDUAL_PTR::const_iterator argIter = argPopul._inds.begin ();
		_inds.reserve (argPopul.getSize());
		for (; iter != end; iter++, argIter++) {
			(*iter)->assign ((*argIter));
		}

	} else {
		if ((int)_inds.size () > 0) {
			clearPopulation ();
		}

		__VECTOR_INDIVIDUAL_PTR::const_iterator iter = argPopul._inds.begin (), end = argPopul._inds.end();
		_inds.reserve (argPopul.getSize());
		for (; iter != end; iter++) {
			_inds.push_back ((*iter)->clone ());
		}
	}

	return (*this);
}

int QPopulation::assign (const QPopulation* argPopul) {
	this->operator = ((QPopulation&)(*argPopul));
	return Q_OK;
}

int QPopulation::getSize () const {
	return (int)_inds.size ();
}

int QPopulation::clearPopulation () {
	__VECTOR_INDIVIDUAL_PTR::iterator iter = _inds.begin(), end = _inds.end ();

	if ((int)_inds.size() % 2) {	// if _inds.size is odd
		delete (*iter);
		iter++;
	}

	for (; iter != end; iter++) {
		delete (*iter);
		iter++;
		delete (*iter);
	}

	_inds.clear ();

	return Q_OK;
}

int QPopulation::checkConvergence () {
	QIndividual* tempInd = _inds[0];
	__VECTOR_INDIVIDUAL_PTR::iterator iter = _inds.begin() + 1, end = _inds.end();
//	_popul->printPopulation (_fout);
	for (; iter != end; iter++) {
		if (! tempInd->equalsTo (*iter) ) {
			return Q_OK;
		}
	}

	return Q_CONVERGENCE;
}

int QPopulation::setParameters (const QEAParameters& argParams) {
	_params = &argParams;

	// create individuals
	if (createPopulation (_params->populationSize) == Q_OK) {
		// create chromosomes
		__VECTOR_INDIVIDUAL_PTR::iterator iter = _inds.begin (), end = _inds.end();
		for (; iter != end; iter++) {
			(*iter)->setParameters (_params);
		}
	}

	return Q_OK;
}

QIndividual* QPopulation::getIndividual (const int argIndex) {
	if (argIndex < (int)_inds.size ()) {
		return _inds[ argIndex ];
	} else {
		cerr<<"\n\n[QPopulation::getIndividual] error:\n"
			<<"\tIndex is out of population bounds (i = "<<argIndex<<")\n\n";
		exit (0);
	}
}

double QPopulation::getIndividualFitnessValue (const int argIndex) const {
	return _inds[argIndex]->getFitnessValue ();
}

QIndividual* QPopulation::getMinFitnessIndividual (const double argBound) {
	double minValue;
	QIndividual* tempInd = NULL;
	__VECTOR_INDIVIDUAL_PTR::iterator iter = _inds.begin (), end = _inds.end();

	minValue = 1000000000;
	for (; iter != end; iter++) {
		if (minValue > (*iter)->getFitnessValue () && minValue > argBound) {
			minValue = (*iter)->getFitnessValue ();
			tempInd = (*iter);
		}
	}

	return tempInd;
}

int QPopulation::getMinFitnessIndividualIndex (const double argBound) const {
	double minValue;
	int index = -1;
	int i=0;
	__VECTOR_INDIVIDUAL_PTR::const_iterator iter = _inds.begin (), end = _inds.end();

	minValue = 1000000000;
	for (; iter != end; iter++) {
		if (minValue > (*iter)->getFitnessValue () && minValue > argBound) {
			minValue = (*iter)->getFitnessValue ();
			index = (int)(iter - _inds.begin());
		}
	}

	return index;
}

const QIndividual* QPopulation::getIndividual (const int argIndex) const {
	if (argIndex < (int)_inds.size ()) {
		return _inds[ argIndex ];
	} else {
		cerr<<"\n\n[QPopulation::getIndividual] error:\n"
			<<"\tIndex is out of population bounds (i = "<<argIndex<<")\n\n";
		exit (0);
	}
}

int QPopulation::setIndividual (const int argIndex, const QIndividual* argInd) {
	if (argIndex < (int)_inds.size ()) {
		_inds[ argIndex ]->assign (argInd);
		return Q_OK;
	} else {
		cerr<<"\n\n[QPopulation::setIndividual] error:\n"
			<<"\tIndex is out of population bounds (i = "<<argIndex<<")\n\n";
		exit (0);
	}
}

int QPopulation::insertIndividual (const int pos, const QIndividual* argInd) {
	_inds.insert (_inds.begin()+pos, argInd->clone());
	return (int)_inds.size ();
}

int QPopulation::addIndividual (const QIndividual* argInd) {
	_inds.push_back (argInd->clone ());
	return (int)_inds.size ();
}

int QPopulation::deleteIndividual (const int argIndex) {
	if (argIndex < (int)_inds.size () && argIndex >=0 ) {

		delete _inds[ argIndex ];
		_inds.erase (_inds.begin() + argIndex);
		return (int)_inds.size ();

	} else {

		cerr<<"\n\n[QPopulation::deleteIndividual] error:\n"
			<<"\tIndex ("<<argIndex<<") is out of population bounds\n\n";
		exit (0);

	}
}

int QPopulation::deleteIndividuals (const int argStart, const int argNumber) {
	if (argStart < (int)_inds.size () && argStart >= 0) {

		int i;
		int number = argNumber;

		if ((argStart + number) >= (int)_inds.size()) {
			number = (int)_inds.size() - argStart;
		}
		for (i=argStart; i<(argStart+number); i++) {
			delete _inds[ i ];
		}
		_inds.erase (_inds.begin() + argStart, _inds.begin() + argStart + number);
		return (int)_inds.size ();

	} else {

		cerr<<"\n\n[QPopulation::deleteIndividuals] error:\n"
			<<"\tIndex ("<<argStart<<") is out of population bounds\n\n";
		exit (0);

	}
}

void QPopulation::reserve (const int argSize) {
	_inds.reserve (argSize);
}

int QPopulation::printPopulation (ostream& out)  const {
	__VECTOR_INDIVIDUAL_PTR::const_iterator iter = _inds.begin (), end = _inds.end();

	out<<"\nPrinting population ("<<(int)(_inds.size())<<" individuals):\n"
		<<"-------------------------\n";
	for (int i=1; iter != end; iter++, i++) {
		out<<"Individual "<<i<<":\t";
		(*iter)->print2Stream (out);
		out<<"\n";
	}

	return Q_OK;
}

