/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qrngene.h"

#include "../include/qec.h"

#include <math.h>

using namespace ECWorkshop;

QRNGene::QRNGene (void) : _bias(0), _enabled(true), _a(1.0), _type(QRNT_SIGMOID) {
}

QRNGene::QRNGene (const QRNGene& argGene)
: QGene(argGene), _input(argGene._input), _output(argGene._output),
_weight(argGene._weight), _weightOuts(argGene._weightOuts), _bias(argGene._bias),
_type(argGene._type), _enabled(true), _a(argGene._a) {
}

QRNGene::~QRNGene (void) {
	_input.clear ();
	_output.clear ();
	_weight.clear ();
	_weightOuts.clear ();
}

QRNGene& QRNGene::operator= (const QRNGene& argGene) {
	this->QGene::operator = (argGene);

	_input.assign (argGene._input.begin(), argGene._input.end());
	_output.assign (argGene._output.begin(), argGene._output.end());
	_weight.assign (argGene._weight.begin(), argGene._weight.end());
	_weightOuts.assign (argGene._weightOuts.begin(), argGene._weightOuts.end());

	_enabled = argGene._enabled;
	_type = argGene._type;
	_bias = argGene._bias;
	_a = argGene._a;

	return (*this);
}

QGene* QRNGene::clone () const {
	return new QRNGene (*this);
}

QEncodingType QRNGene::getEncodingType () const {
	return QET_REGULATORY_NETWORK;
}

bool QRNGene::equalsTo (const QGene* argGene) const {
	if (argGene->getEncodingType() == QET_REGULATORY_NETWORK) {
		QRNGene* tempGene = (QRNGene*)argGene;
		if (_input.size() == tempGene->_input.size() &&
			_output.size() == tempGene->_output.size() &&
			_weight.size() == tempGene->_weight.size() &&
			_weightOuts.size() == tempGene->_weightOuts.size()) {

			int size = (int)_input.size();
			int i;

			// check input proteins and their weights
			for (i=0; i<size; i++) {
				if (_input[i] != tempGene->_input[i] || _weight[i] != tempGene->_weight[i]) {
					return false;
				}
			}

			// check product proteins and their weights
			size = (int)_output.size();
			for (i=0; i<size; i++) {
				if (_output[i] != tempGene->_output[i] ||  _weightOuts[i] != tempGene->_weightOuts[i]) {
					return false;
				}
			}

			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}	// if (argGene->getEncodingType() == QET_REGULATORY_NETWORK)
}

int QRNGene::assign (const QGene* argGene) {
	if (argGene->getEncodingType() == QET_REGULATORY_NETWORK) {
		QRNGene* tempGene = (QRNGene*)argGene;

		_input.assign (tempGene->_input.begin(), tempGene->_input.end());
		_output.assign (tempGene->_output.begin(), tempGene->_output.end());
		_weight.assign (tempGene->_weight.begin(), tempGene->_weight.end());
		_weightOuts.assign (tempGene->_weightOuts.begin(), tempGene->_weightOuts.end());

		_enabled = tempGene->_enabled;
		_bias = tempGene->_bias;
		_type = tempGene->_type;
		_a = tempGene->_a;

		return Q_OK;
	} else {
		return Q_FAILURE;
	}
}

QSignal* QRNGene::getOutput (const int argIndex) {
	return &_output[argIndex];
}

const QSignal* QRNGene::getOutput (const int argIndex) const {
	return &_output[argIndex];
}

QSignal* QRNGene::getOutputByID (const int argID) {
	__VECTOR_SIGNAL::iterator iter = _output.begin(), iEnd = _output.end();

	for (; iter != (iEnd); iter++) {
		if ((*iter)._id == argID) {
			return &(*iter);
		}
	}

	return NULL;
}

QSignal* QRNGene::getInput (const int argIndex) {
	return &_input[argIndex];
}

const QSignal* QRNGene::getInput (const int argIndex) const {
	return &_input[argIndex];
}

int QRNGene::setInput (const QSignal& argValue) {
	return this->setInputByID (argValue._id, argValue._amount);
}

int QRNGene::setInputByID (const int argID, const double argAmount) {
	__VECTOR_SIGNAL::iterator iter = _input.begin(), iEnd = _input.end();

	for (; iter != (iEnd); iter++) {
		if ((*iter)._id == argID) {
			(*iter)._amount = argAmount;
			return Q_OK;
		}
	}

	return Q_FAILURE;
}

int QRNGene::setInput (const int argIndex, const double argAmount) {
	_input[argIndex]._amount = argAmount;
	return Q_OK;
}

int QRNGene::setInput (const int argIndex, const QSignal& argValue) {
	_input[argIndex] = argValue;
	return Q_OK;
}

int QRNGene::setOutput (const int argIndex, const QSignal& argValue) {
	_output[argIndex] = argValue;
	return Q_OK;
}

int QRNGene::setInputs (const __VECTOR_SIGNAL& argIns) {
	_input.assign (argIns.begin(), argIns.end());
	_weight.resize (_input.size());
	return Q_OK;
}

int QRNGene::setOutputs (const __VECTOR_SIGNAL& argOuts) {
	_output.assign (argOuts.begin(), argOuts.end());
	_weightOuts.resize (_output.size());
	return Q_OK;
}

int QRNGene::calculate (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState) {
	if (!_enabled) {
		return Q_FAILURE;
	}
	int size = (int)_input.size ();
	int i, stateSize = (int)curState.size (), tempID;
	double sum, output = 0.0;

	// calculate weighted sum of input signals
	sum = _bias;
	for (i=0; i<size; i++) {
		tempID = _input[i]._id;
		if (tempID != 0) {	// if the i-th input is enabled
			if (stateSize > tempID) {
				_input[i]._amount = curState[tempID];
			} else {
				_input[i]._amount = 0;
			}
			sum += _weight[i] * _input[i]._amount;
		}
	}

	// calculate overall node's output signal using one of available functions
	double temp;
	sum *= _a;
	switch (_type) {
		case QRNT_LINEAR:
			if (sum < 1 && sum > (-1)) {
				output = sum;
			} else if (sum > 1) {
				output = 1;
			} else {	// sum < -1
				output = -1;
			}
			break;
		case QRNT_SIGMOID:
			temp = exp(-2*sum);
			output = (1.0-temp)/(1.0+temp);
			break;
		case QRNT_THRESHOLD:
			if (sum > 0) {
				output = 1;
			} else {
				output = -1;
			}
			break;
		case QRNT_SIN:
			output = sin (sum);
			break;
		default:
			cout<<"\n\n[QRNGene::calculate] error:\n"
				<<"Unknown node type (typeID = "<<_type<<")\n";
			break;
	}

	// distribute overall output signal along each kind of output signals
	// and update the newState vector

	size = (int)_output.size();
	for (i=0; i<size; i++) {
		tempID = _output[i]._id;
		if (tempID != 0) {
			_output[i]._amount = _weightOuts[i] * output;
			while ((int)newState.size() <= tempID) {
				newState.push_back (0);
			}
			newState [tempID] += _output[i]._amount;
		}
	}

	return Q_OK;
}

bool QRNGene::isEnabled () const {
	return _enabled;
}

int QRNGene::enable (bool argStatus) {
	_enabled = argStatus;
	return Q_OK;
}

int QRNGene::getInputsNumber () const {
	return (int)_input.size ();
}

int QRNGene::setInputsNumber (const int argN) {
	_input.resize (argN);
	_weight.resize (argN);
	return (int)_input.size();
}

int QRNGene::getOutputsNumber () const {
	return (int)_output.size ();
}

int QRNGene::setOutputsNumber (const int argN) {
	_output.resize (argN);
	_weightOuts.resize (argN);
	return (int)_output.size();
}

// set weight for the desired input signal
int QRNGene::setInputWeight (const int argIndex, double argW) {
	_weight [argIndex] = argW;
	return Q_OK;
}

// set weight for the desired output signal
int QRNGene::setOutputWeight (const int argIndex, double argW) {
	_weightOuts [argIndex] = argW;
	return Q_OK;
}

int QRNGene::print2Stream (ostream& out) const {
	int i;

	out<<"\tType ID:\t"<<_type<<"\n"
		<<"\tEnabled:\t"<<_enabled<<"\n"
		<<"\tBias:\t"<<_bias<<"\n"
		<<"\tA:\t"<<_a<<"\n"
		<<"\tInputs ("<<(int)_input.size()<<"):\t";
	for (i=0; i<(int)_input.size(); i++) {
		out<<"("<<_input[i]._id<<", "<<_input[i]._amount<<":"<<_weight[i]<<")\t";
	}
	out<<"\n\tOutputs ("<<(int)_output.size()<<"):\t";
	for (i=0; i<(int)_output.size(); i++) {
		out<<"("<<_output[i]._id<<", "<<_output[i]._amount<<":"<<_weightOuts[i]<<")\t";
	}
	out<<"\n";
	return Q_OK;
}

double QRNGene::getBias () const {
	return _bias;
}

int QRNGene::setBias (const double argB) {
	_bias = argB;
	return Q_OK;
}

int QRNGene::addInputSignal (const QSignal* argS, const double argW) {
	_input.push_back (*argS);
	_weight.push_back (argW);
	return Q_OK;
}

int QRNGene::addOutputSignal (const QSignal* argS, const double argW) {
	_output.push_back (*argS);
	_weightOuts.push_back (argW);
	return Q_OK;
}

int QRNGene::deleteInputSignalByID (const int argID) {
	__VECTOR_SIGNAL::iterator iter = _input.begin();

	for (; iter != _input.end(); ) {
		if ((*iter)._id == argID) {
			if (iter != _input.begin()) {
				iter --;
				_input.erase (iter+1);
				_weight.erase (_weight.begin() + (int)(iter - _input.begin() + 1));
				iter ++;
			} else {
				_input.erase (iter);
				_weight.erase (_weight.begin());
				iter = _input.begin ();
			}
		} else {
			iter++;
		}
	}

	return Q_OK;
}

int QRNGene::deleteOutputSignalByID (const int argID) {
	__VECTOR_SIGNAL::iterator iter = _output.begin();

	for (; iter != _output.end(); ) {
		if ((*iter)._id == argID) {
			if (iter != _output.begin()) {
				iter --;
				_output.erase (iter+1);
				_weightOuts.erase (_weightOuts.begin() + (int)(iter - _output.begin() + 1));
				iter ++;
			} else {
				_output.erase (iter);
				_weightOuts.erase (_weightOuts.begin());
				iter = _output.begin ();
			}
		} else {
			iter++;
		}
	}

	return Q_OK;
}

bool QRNGene::hasInput (const int id) {
	__VECTOR_SIGNAL::iterator iter = _input.begin();

	for (; iter != _input.end(); iter++) {
		if ((*iter)._id == id) {
			return true;
		}
	}

	return false;
}

bool QRNGene::hasOutput (const int id) {
	__VECTOR_SIGNAL::iterator iter = _output.begin();

	for (; iter != _output.end(); iter++) {
		if ((*iter)._id == id) {
			return true;
		}
	}

	return false;
}

QRNNodeType QRNGene::getType () const {
	return _type;
}

int QRNGene::setType (const QRNNodeType argT) {
	_type = argT;
	return Q_OK;
}

double QRNGene::getParamA () const {
	return _a;
}

int QRNGene::setParamA (const double argA) {
	_a = argA;
	return Q_OK;
}

int QRNGene::changeInputID (const int oldID, const int newID) {
	__VECTOR_SIGNAL::iterator iter = _input.begin();

	for (; iter != _input.end(); iter++) {
		if ((*iter)._id == oldID) {
			(*iter)._id = newID;
		}
	}

	return Q_OK;
}

int QRNGene::changeOutputID (const int oldID, const int newID) {
	__VECTOR_SIGNAL::iterator iter = _output.begin();

	for (; iter != _output.end(); iter++) {
		if ((*iter)._id == oldID) {
			(*iter)._id = newID;
		}
	}

	return Q_OK;
}

double QRNGene::getInputWeight (const int argIndex) const {
	return _weight[argIndex];
}

double QRNGene::getOutputWeight (const int argIndex) const {
	return _weightOuts[argIndex];
}

int QRNGene::getMaxInputID () const {
	__VECTOR_SIGNAL::const_iterator iter = _input.begin(), iEnd = _input.end ();
	int maxID;

	maxID = 0;
	for (; iter != iEnd; iter++) {
		if ((*iter)._id > maxID) {
			maxID = (*iter)._id;
		}
	}

	return maxID;
}

int QRNGene::getMaxOutputID () const {
	__VECTOR_SIGNAL::const_iterator iter = _output.begin(), iEnd = _output.end ();
	int maxID;

	maxID = 0;
	for (; iter != iEnd; ++iter) {
		if ((*iter)._id > maxID) {
			maxID = (*iter)._id;
		}
	}

	return maxID;
}
