/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qrnindividual.h"

#include "../operators/mutation/qrnmutation.h"

#include <math.h>

using namespace ECWorkshop;

int QRNIndividual::createIndividual (const int argGenesNumber) {
//	_genes.resize (argGenesNumber);
	_genes.resize (_params->outputs);
	_mParams.resize (MUTATION_PARAMETERS);
	_history.assign (MUTATION_PARAMETERS, 0);	// init history array
	_inputsNumber = _params->inputs;
	_outputsNumber = _params->outputs;
	return Q_OK;
}

QRNIndividual::QRNIndividual (void)
: _inputsNumber(0), _outputsNumber(0), _energy(0) {
	_history.assign (MUTATION_PARAMETERS, 0);	// init history array
	_mParams.resize (MUTATION_PARAMETERS);
}

QRNIndividual::QRNIndividual (const QRNIndividual& argInd)
: QIndividual(argInd), _inputsNumber(argInd._inputsNumber),
_outputsNumber(argInd._outputsNumber), _energy(argInd._energy) {
	_genes.assign (argInd._genes.begin(), argInd._genes.end());
	_mParams.assign (argInd._mParams.begin(), argInd._mParams.end());
	_history.assign (argInd._history.begin (), argInd._history.end ());
}

QRNIndividual::~QRNIndividual(void) {
	_genes.clear ();
	_mParams.clear ();
	_history.clear ();
}

QRNIndividual& QRNIndividual::operator= (const QRNIndividual& argInd) {
	this->QIndividual::operator = (argInd);
	_genes.assign (argInd._genes.begin(), argInd._genes.end());
	_mParams.assign (argInd._mParams.begin(), argInd._mParams.end());
	_history.assign (argInd._history.begin (), argInd._history.end ());
	_energy = argInd._energy;
	_inputsNumber = argInd._inputsNumber;
	_outputsNumber = argInd._outputsNumber;
	return (*this);
}

int QRNIndividual::assign (const QIndividual* argInd) {
	this->operator= (*(QRNIndividual*)argInd);
	return Q_OK;
}

int QRNIndividual::getSize () const {
	return (int)_genes.size ();
}

int QRNIndividual::addGene (const QGene* argGene) {
	_genes.push_back (*(QRNGene*)argGene);
	return (int)_genes.size ();
}

int QRNIndividual::deleteGene (const int argIndex) {
	if (argIndex < (int)_genes.size ()) {

		_genes.erase (_genes.begin() + argIndex);
		return (int)_genes.size ();

	} else {

		cerr<<"\n\n[QRNIndividual::deleteGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);

	}
}

QGene* QRNIndividual::getGene (const int argIndex) {
	if (argIndex < (int)_genes.size ()) {
		return &_genes[ argIndex ];
	} else {
		cerr<<"\n\n[QRNIndividual::getGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

const QGene* QRNIndividual::getGene (const int argIndex) const {
	if (argIndex < (int)_genes.size ()) {
		return &_genes[ argIndex ];
	} else {
		cerr<<"\n\n[QRNIndividual::getGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

int QRNIndividual::setGene (const int argIndex, const QGene* argGene) {
	if (argIndex < (int)_genes.size ()) {
		_genes[ argIndex ].assign (argGene);
		return Q_OK;
	} else {
		cerr<<"\n\n[QRNIndividual::setGene] error:\n"
			<<"\tIndex is out of genes array bounds\n\n";
		exit (0);
	}
}

int QRNIndividual::clearIndividual () {
	_genes.clear ();
	return Q_OK;
}

// get value of mutation parameter with specified index
double QRNIndividual::getMutationParam (const int argIndex) {
	return _mParams[argIndex];
}

// set value of mutation parameter with specified index
int QRNIndividual::setMutationParam (const int argIndex, const double argValue) {
	_mParams[argIndex] = argValue;
	return Q_OK;
}

bool QRNIndividual::equalsTo (const QIndividual* argInd) const {
	if (getEncodingType() == argInd->getEncodingType()) {
		const QRNIndividual* tempInd = (QRNIndividual*)argInd;
		if (_genes.size() == tempInd->_genes.size()) {
			__VECTOR_RN_GENE::const_iterator iter = _genes.begin (), end = _genes.end();
			__VECTOR_RN_GENE::const_iterator argIter = tempInd->_genes.begin ();

			for (; iter != end; iter++, argIter++) {
				if ((*iter).equalsTo (&(*argIter)) == false) {
					return false;
				}
			}
			return true;
		}
	}

	return false;
}

QIndividual* QRNIndividual::clone () const {
	return new QRNIndividual (*this);
}

QEncodingType QRNIndividual::getEncodingType () const {
	return QET_REGULATORY_NETWORK;
}

//[type] QRNIndividual::getGeneValue (const int argIndex) const {
//	return _genes[ argIndex ].getValue ();
//}
//
//int QRNIndividual::setGeneValue (const int argIndex, const [type] argValue) {
//	if (argIndex < (int)_genes.size () && argIndex >= 0) {
//		_genes[ argIndex ].setValue (argValue);
//		return Q_OK;
//	} else {
//		cerr<<"\n\nQRNIndividual::setGeneValue error:\n"
//			<<"\tIndex is out of genes array bounds\n\n";
//		exit (0);
//	}
//}

int QRNIndividual::print2Stream (ostream& out) const {
	__VECTOR_RN_GENE::const_iterator iter = _genes.begin (), end = _genes.end ();
	const QRNGene *tempGene = NULL;

	out<<"(";
	_fitness.print2Stream (out);
	out<<")\t"<<_inputsNumber<<" inputs : "<<_outputsNumber<<" outputs\n"
		<<"Genes number:\t"<<_genes.size()<<"\n";
	//while (iter != end) {
	//	out<<"Gene "<<(int)(iter-_genes.begin())<<":";
	//	(*iter).print2Stream (out);
	//	++iter;
	//}
	//if (!_genes.empty()) {
	for (; iter != end; iter++) {
		out<<"Gene "<<(int)(iter-_genes.begin())<<":";
		(*iter).print2Stream (out);
	}
	out<<"Mutation parameters:\t";
	for (int i=0; i<MUTATION_PARAMETERS; ++i) {
		out<<" "<<_mParams[i];
	}
	out<<"\n";
	out<<"Total number of signals:\t"<<(max(getMaxInputID(), getMaxOutputID()))<<"\n";

	// Print individual's history
	out<<"History events counters:\t";
	for (int i=0; i<MUTATION_PARAMETERS; ++i) {
		out<<" "<<_history[i];
	}
	out<<"\n";

	// Print average degrees for nodes
	__VECTOR_INT tempV;
	double avDegr;
	calculateDegrees (tempV);
	avDegr = qavg (tempV);
	out<<"Average degree:\t"<<avDegr<<"\n";
	out<<"Nodes degrees:\t";
	for (int i=0; i<(int)tempV.size(); ++i) {
		out<<" "<<tempV[i];
	}
	out<<"\n";

	// print signals degrees
	__VECTOR_INT dIn, dOut;
	calculateDegreesSignals (dIn, dOut);
	out<<"Input signals degrees:\t";
	for (int i=0; i<(int)dIn.size(); ++i) {
		out<<" "<<dIn[i];
	}
	out<<"\n";
	out<<"Output signals degrees:\t";
	for (int i=0; i<(int)dOut.size(); ++i) {
		out<<" "<<dOut[i];
	}
	out<<"\n";


	//
	// PROBLEM: How to calculate the clustering coefficient taking into consideration multiple connections
	//		betwenn nodes?
	//

	// print clustering coefficient
//	double cCoef = 0;
//	int size = tempV.size (), tempSum = 0;
//	for (int i=0; i<size; ++i) {
//		if (tempV[i] > 1) {
////			cCoef += 2.0/((double)tempV[i] * (tempV[i] - 1.0));
//			cCoef += 1.0/((double)(tempV[i] - 1.0));	// the factor 1.0 due to use of directed connections
//		}
//	}
//	cCoef /= (double)size;
//	out<<"Clustering coefficient:\t"<<cCoef<<"\n";
//	
//	// print characteristic path-length
//	double pathL;
//	pathL = log((double)_genes.size())/log(avDegr);
//	out<<"Characteristic path-length:\t~"<<pathL<<"\n";
//
//	// calculate characterisitcs for random graph
//	__VECTOR_INT dIn, dOut;
//	calculateDegreesSignals (dIn, dOut);
//	out<<"Clustering coefficient (random graph):\t"<<cCoef<<"\n";

	//}
	return Q_OK;
}

int QRNIndividual::toVectorDouble (__VECTOR_DOUBLE& target) {
	// Convert genes array to the array of double if possible and
	//  write result into the [target] variable
	// Return Q_OK on success or Q_FAILURE if the operation is
	//  unsupported or error encountered
	return Q_FAILURE;
}

int QRNIndividual::toVectorDouble (__VECTOR_DOUBLE& target) const {
	// Constant variant of the previous method
	// Convert genes array to the array of double if possible and
	//  write result into the [target] variable
	// Return Q_OK on success or Q_FAILURE if the operation is
	//  unsupported or error encountered
	return Q_FAILURE;
}

int QRNIndividual::toVectorInt (__VECTOR_INT& target) {
	// Convert genes array to the array of int if possible and
	//  write result into the [target] variable
	// Return Q_OK on success or Q_FAILURE if the operation is
	//  unsupported or error encountered
	return Q_FAILURE;
}

int QRNIndividual::toVectorInt (__VECTOR_INT& target) const {
	// Constant variant of the previous method
	// Convert genes array to the array of int if possible and
	//  write result into the [target] variable
	// Return Q_OK on success or Q_FAILURE if the operation is
	//  unsupported or error encountered
	return Q_FAILURE;
}

int QRNIndividual::calculate (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState, bool refreshInputs) {
	int maxDist,	// maximum distance from input to output
		minDist;	// minimum distance from input to output
	int i, j;
	__VECTOR_DOUBLE tempState;

	// TODO: Define number of update steps
	calculateMinMaxDistances (minDist, maxDist);

	if (minDist > 0) {
		// Make updates for [maxDist] number of steps
		tempState.assign (curState.begin(), curState.end());
		for (i=0; i < maxDist; ++i) {
			// Clear the new state (it is assumed that all the signals were transferred)
			newState.clear ();
			calculate1step (tempState, newState);
			// update states
			tempState.assign (newState.begin(), newState.end());
			while ((int)tempState.size() < _inputsNumber + 1) {
				tempState.push_back (0);
			}
			// assign input signals' amounts
			if (!refreshInputs) {
				for (j=1; j<=_inputsNumber; ++j) {
					tempState[j] = curState[j];
				}
			}
		}
	} else {
		// fill [newState] with zeroes
		newState.assign (_inputsNumber + _outputsNumber + 1, 0);
	}	// if (minDist > 0)

	return Q_OK;
}

int QRNIndividual::calculate1step (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState) {
	__VECTOR_RN_GENE::iterator iter = _genes.begin (), end = _genes.end ();

	for (; iter != end; iter++) {
		(*iter).calculate (curState, newState);
	}

	return Q_OK;
}

int QRNIndividual::calculateNsteps (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState, const int argSteps, bool refreshInputs) {
	int i, j, size, prevSize;
	__VECTOR_DOUBLE tempState;

	tempState.assign (curState.begin(), curState.end());
	_energy = 0;
	for (i=0; i<argSteps; ++i) {
		calculate1step (tempState, newState);
		//calculate energy by change of the signals power
		size = (int)newState.size ();
		prevSize = (int)tempState.size ();
		for (j=_inputsNumber+1; j<size; ++j) {	// calculate the difference for all non-input signals only
			if (j < prevSize) {
				_energy += fabs (newState[j] - tempState[j]);
			} else {
				_energy += fabs (newState[j]);
			}
		}

		// update environment state
		tempState.assign (newState.begin(), newState.end());
		while ((int)tempState.size() < _inputsNumber + 1) {
			tempState.push_back (0);
		}
		if (!refreshInputs) {
			for (j=1; j<=_inputsNumber; ++j) {
				tempState[j] = curState[j];
			}
		}
	}
	_energy /= argSteps;

	return Q_OK;
}

// calculate degrees for each node
int QRNIndividual::calculateDegrees (__VECTOR_INT& argD) const {
	int i, j, size = (int)_genes.size();
	int ins, outs, tempID;
	__VECTOR_INT tempI, tempO;	// arrays to store degrees
								// for input and output signals usage in nodes

	// calculate degrees for inputs and outputs and 
	// store them into the [tempI] and [tempO] arrays
	calculateDegreesSignals (tempI, tempO);
	//for (i=1; i<=_inputsNumber; i++) {
	//	while ((int)tempO.size () <= i) {
	//		tempO.push_back (0);
	//	}
	//	tempO[i] = 1;	// take into consideration input signals for the entire RN
	//}
	//for (i=_inputsNumber+1; i<=_inputsNumber+_outputsNumber; i++) {
	//	while ((int)tempI.size () <= i) {
	//		tempI.push_back (0);
	//	}
	//	tempI[i] = 1;	// take into consideration output signals for the entire RN
	//}

	//ins = (int)tempI.size ();
	//for (i=0; i<ins; i++) {
	//	cout<<tempI[i]<<", ";
	//}
	//cout<<"\n";
	//outs = (int)tempO.size ();
	//for (i=0; i<outs; i++) {
	//	cout<<tempO[i]<<", ";
	//}
	//cout<<"\n";

	argD.resize (size);
	argD.assign (size, 0);
	for (i=0; i<size; i++) {
//		cout<<argD[i]<<"\t";
		ins = _genes[i].getInputsNumber();
		outs = _genes[i].getOutputsNumber();

		// calculate degree for the i-th node's inputs
		for (j=0; j<ins; j++) {
			tempID = _genes[i].getInput(j)->_id;
			if (tempID < (int)tempO.size ()) {
				argD[ i ] += tempO [ tempID ];
			}	// otherwise do not increment argD[i]
		}

		// calculate degree for the i-th node's outputs
		for (j=0; j<outs; j++) {
			tempID = _genes[i].getOutput(j)->_id;
			if (tempID < (int)tempI.size()) {
				argD[ i ] += tempI [ tempID ];
			}	// otherwise do not increment argD[i]
		}
	}	// for (i=0; i<size; i++)

	return Q_OK;
}

// calculate degrees for input and output signals
int QRNIndividual::calculateDegreesSignals (__VECTOR_INT& argIn, __VECTOR_INT& argOut) const {
	int i, j, size = (int)_genes.size();
	int inputs, outputs, tempID;
	QRNGene* tempGene;

	argIn.push_back (0);
	argOut.push_back (0);
	// calculate degrees using only information about network
	for (i=0; i<size; ++i) {
		tempGene = &_genes[i];

		// calculate input signals degrees
		inputs = tempGene->getInputsNumber();
		for (j=0; j<inputs; ++j) {
			tempID = tempGene->getInput(j)->_id;
			while (tempID>= (int)(argIn.size())) {
				argIn.push_back (0);
			}
			++argIn[tempID];
		}

		// calculate output signals degrees
		outputs = tempGene->getOutputsNumber();
		for (j=0; j<outputs; ++j) {
			tempID = tempGene->getOutput(j)->_id;
			while (tempID>= (int)(argOut.size())) {
				argOut.push_back (0);
			}
			++argOut[tempID];
		}
	}	// for (i=0; i<size; i++)

	// update degrees taking into consideration pseudo inputs and outputs
	for (i=1; i<=_inputsNumber; ++i) {
		while ((int)argOut.size () <= i) {
			argOut.push_back (0);
		}
		argOut[i] += 1;	// take into consideration input signals for the entire RN
	}
	for (i=_inputsNumber+1; i<=_inputsNumber+_outputsNumber; ++i) {
		while ((int)argIn.size () <= i) {
			argIn.push_back (0);
		}
		argIn[i] += 1;	// take into consideration output signals for the entire RN
	}

	// null the 0-th input/output signals' degree
	argIn[0] = argOut[0] = 0;

	return Q_OK;
}

// calculates min and max distances from input to output using connectivity matrix
// returns Q_OK on success
int QRNIndividual::calculateMinMaxDistances (int& minDist, int& maxDist) const {
	int i, j, inputs, outputs, tempID, size;//, stepsLimit;
	__TABLE_BYTE cMatrix, tempM, tempM_1;
	__VECTOR_INT tempV;
	__VECTOR_RN_GENE::const_iterator iter = _genes.begin(), iEnd = _genes.end();

	//////////////////////////////////////////
	// Init the connectivity matrix [cMatrix]
	//
	//	TODO:	Merge inputs and outputs to reduce the matrix dimensions
	//
	size = max (getMaxInputID (), getMaxOutputID ());
	//cout<<"getMaxInputID = "<<getMaxInputID()<<"\n";
	//cout<<"getMaxOutputID = "<<getMaxOutputID()<<"\n";
	//cout<<"size = "<<size<<"\n";
	cMatrix.resize (size);
	for (i=0; i<size; ++i) {
		cMatrix[i].assign (size, 0);
	}

	// cycle through all nodes
	for (; iter != iEnd; ++iter) {
		if ((*iter).isEnabled()) {
			// prepare temporary array
			tempV.clear ();
			// cycle through all the input signals in the current node
			inputs = (*iter).getInputsNumber ();
			//cout<<"inputs = "<<inputs<<"\n";
			for (i=0; i<inputs; ++i) {
				// collect IDs of input signals
				if ((*iter).getInput(i)->_id > 0) {
					tempV.push_back ((*iter).getInput(i)->_id - 1);
					//cout<<"tempV["<<((int)tempV.size()-1)<<"] = "<<tempV[ (tempV.size()-1) ]<<", ";
					//cout.flush ();
				}
			}
			//cout<<"\n";
			//cout.flush ();

			outputs = (*iter).getOutputsNumber();
			for (i=0; i<outputs; ++i) {
				if ((*iter).getOutput(i)->_id > 0) {
					tempID = (*iter).getOutput(i)->_id - 1;
					// fill the tempID-th column of [cMatrix]
					for (j=0; j<(int)tempV.size(); ++j) {
						cMatrix[ tempV[j] ][ tempID ] = 1;
					}
				}
			}
		}	// if ((*iter).isEnabled())
	}	// for (; iter != iEnd; ++iter)

	//cout<<"cMatrix is filled\n";
	//cout.flush ();

	// reset variables
	maxDist = 0;
	minDist = 0;

	// early check: if there are no non-zero values in the columns from the
	// range [_inputsNumber; _inputsNumber+_outputsNumber-1] then the output is
	// unreachable (minDist = 0)
	if (size < (_inputsNumber+1)) {	// no signals are processed inside the RN in fact
		return Q_FAILURE;
	}
	//for (i=0; i<size; ++i) {
	//	for (j=_inputsNumber; j<_inputsNumber+_outputsNumber && j < size; ++j) {
	//		if (cMatrix[i][j] != 0) {
	//			break;
	//		}
	//	}
	//	if (j < (_inputsNumber+_outputsNumber)) {
	//		// if non-zero value is found
	//		break;
	//	}
	//}
	//if (i == size) {
	//	// if no non-zero values are found (only zeroes are present) then
	//	// minDist = maxDist = 0
	//	return Q_FAILURE;
	//}

	// early check 2: if there are no non-zero values in the rows from the
	// range [0; _inputsNumber-1] then the input is unreachable (minDist = 0)
	for (i=0; i<_inputsNumber; ++i) {
		for (j=0; j<size; ++j) {
			if (cMatrix[i][j] != 0) {
				break;
			}
		}
		if (j < size) {
			// if non-zero value is found
			break;
		}
	}
	if (i == _inputsNumber) {
		// if no non-zero values are found (only zeroes are present) then
		// minDist = maxDist = 0
		return Q_FAILURE;
	}

	// Calculate [minDist] and [maxDist]
	// by continuous "powerization" of [cMatrix]
	tempM.assign (cMatrix.begin(), cMatrix.end());
	while (! isNull (tempM)) {
		// check whether there are all-zero values in the columns from the
		// range [_inputsNumber; _inputsNumber+_outputsNumber-1]. If so then the output is
		// unreachable (minDist = 0)
		for (i=0; i<size; ++i) {
			for (j=_inputsNumber; j<_inputsNumber+_outputsNumber && j < size; ++j) {
				if (tempM[i][j] != 0) {
					break;
				}
			}
			if (j < (_inputsNumber+_outputsNumber)) {
				// if non-zero value is found
				break;
			}
		}
		if (i == size) {
			// if no non-zero values are found (only zeroes are present) then
			// minDist = maxDist = 0
			return Q_FAILURE;
		}

		// early check 2: if there are no non-zero values in the rows from the
		// range [0; _inputsNumber-1] then the input is unreachable (minDist = 0)
		for (i=0; i<_inputsNumber; ++i) {
			for (j=0; j<size; ++j) {
				if (tempM[i][j] != 0) {
					break;
				}
			}
			if (j < size) {
				// if non-zero value is found
				break;
			}
		}
		if (i == _inputsNumber) {
			// if no non-zero values are found (only zeroes are present) then
			// minDist = maxDist = 0
			return Q_FAILURE;
		}

		// look for the shortest path
		// by checking the 1st [_inputsNumber] rows
		if (minDist == 0) {
			for (i=0; i<_inputsNumber; ++i) {
				for (j=_inputsNumber; j<_inputsNumber+_outputsNumber; ++j) {
					if (tempM[i][j] != 0) {
						minDist = maxDist + 1;
						break;
					}
				}
				if (minDist != 0) {	// if [minDist] is assigned with some value
					break;
				}
			}
		}
		// calculate the next matrix of distance
		tempM_1.assign (tempM.begin(), tempM.end());
		tempM = tempM * cMatrix;
		++maxDist;	// use [maxDist] as a counter
		if (tempM == tempM_1) {
			// the connectivity matrix is stabilized (infinite loop found)
			maxDist *= 2;
			return Q_OK;
		}

		if (maxDist > size) {
			//cout<<"Keep cycling! ("<<minDist<<" : "<<maxDist<<")\n";
//			print2Stream (cout);
//			exit (0);
			return Q_FAILURE;
		}

		// if there are non-zero values on the main diagonal at the [tempM]
		// then there is an infinite loop
		for (i=0; i<size; ++i) {
			if (tempM[i][i] != 0) {
				break;
			}
		}
		if (i < size && minDist != 0) {
			// increase [maxDist] to allow the loop to influence on RN output
			//cout<<"Loop found\n";
			maxDist *= 2;
			return Q_OK;
		}

	}

	return Q_OK;
}

double QRNIndividual::getEnergy () const {
	return _energy;
}

int QRNIndividual::getMaxInputID () const {
	__VECTOR_RN_GENE::const_iterator iter = _genes.begin(), iEnd = _genes.end ();
	int maxID;

	maxID = 0;
	for (; iter != iEnd; ++iter) {
		if ((*iter).getMaxInputID() > maxID) {
			maxID = (*iter).getMaxInputID();
		}
	}

	return maxID;
}

int QRNIndividual::getMaxOutputID () const {
	__VECTOR_RN_GENE::const_iterator iter = _genes.begin(), iEnd = _genes.end ();
	int maxID;

	maxID = 0;
	for (; iter != iEnd; iter++) {
		if ((*iter).getMaxOutputID() > maxID) {
			maxID = (*iter).getMaxOutputID();
		}
	}

	return maxID;
}

// Returns _history[argID] (number of changes with the type ID = argID);
int QRNIndividual::readHistory (const int argID) const {
	return _history[argID];
}

// increments _history[argID] element (number of changes with the type ID = argID);
int QRNIndividual::writeHistory (const int argID) {
	++_history[argID];
	return Q_OK;
}
