/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#ifndef	Q_REGULATORYNETWORK_INDIVIDUAL_H
#define	Q_REGULATORYNETWORK_INDIVIDUAL_H

#include "qindividual.h"

#include "qgene.h"
#include "qrngene.h"

using namespace std;

namespace ECWorkshop {

///////////////////////////////////////////////////////////
//
//	Class QRNIndividual
//
///////////////////////////////////////////////////////////

//#define	MUTATION_PARAMETERS	10

class QRNIndividual : public QIndividual {
protected:

	mutable __VECTOR_RN_GENE _genes;
	// mutation parameters (-1; +1) (by indices):
	// 0 - add/delete RN node
	// 1 - add/delete input for random existing node
	// 2 - add/delete output for random existing node
	// 3 - random change of type for existing node
	// 4 - random change of kind of random input for existing node
	// 5 - random change of kind of random output for existing node
	//     (including ability to generate new kind of signal)
	// 6 - change of weight for random input signal
	// 7 - change of weight for random output signal
	// 8 - toggle random gene on/off
	__VECTOR_DOUBLE _mParams;
	__VECTOR_INT _history;	// stores information about changes with individual's chromosome
	int _inputsNumber, _outputsNumber;	// number of inputs and outputs for the
							// whole network
	double _energy;

public:
	QRNIndividual (void);
	QRNIndividual (const QRNIndividual& argInd);
	virtual ~QRNIndividual(void);

	virtual QRNIndividual& operator= (const QRNIndividual& argInd);
	virtual int assign (const QIndividual* argInd);

	virtual int createIndividual (const int argGenesNumber);

	virtual int getSize () const;
	virtual int addGene (const QGene* argGene);
	virtual int deleteGene (const int argIndex);
	virtual QGene* getGene (const int argIndex);
	virtual const QGene* getGene (const int argIndex) const;
	virtual int setGene (const int argIndex, const QGene* argGene);
	virtual int clearIndividual ();
	virtual bool equalsTo (const QIndividual* argInd) const;
	
	// get value of mutation parameter with specified index
	double getMutationParam (const int argIndex);
	// set value of mutation parameter with specified index
	int setMutationParam (const int argIndex, const double argValue);

	virtual QIndividual* clone () const;

	virtual QEncodingType getEncodingType () const;

	virtual int toVectorDouble (__VECTOR_DOUBLE& target);
	virtual int toVectorInt (__VECTOR_INT& target);
	virtual int toVectorDouble (__VECTOR_DOUBLE& target) const;
	virtual int toVectorInt (__VECTOR_INT& target) const;

	// get max ids for input and output signals
	virtual int getMaxInputID () const;
	virtual int getMaxOutputID () const;

	// Returns _history[argID] (number of changes with the type ID = argID);
	int readHistory (const int argID) const;

	// increments _history[argID] element (number of changes with the type ID = argID);
	int writeHistory (const int argID);

	// Returns average amount of energy consumed during the last [calculateNsteps] method call
	double getEnergy () const;

	// Makes single update of the new state
	int calculate1step (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState);

	// Makes N updates of the new state
	int calculateNsteps (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState, const int argSteps, bool refreshInputs = false);

	// Makes numerous updates of the new state taking into consideration
	//	[RN] structure
	// If [refreshInputs] is <false> then input signals will be updated every cycle of
	//  calculation. Otherwise ([refreshInputs] = true) input signals will be reset to zeroes
	//  in all but the 1st cycle
	int calculate (const __VECTOR_DOUBLE& curState, __VECTOR_DOUBLE& newState, bool refreshInputs = false);

	// calculate degrees for each node
//	virtual int calculateDegrees (__VECTOR_INT& argD);
	virtual int calculateDegrees (__VECTOR_INT& argD) const;

	// calculate degrees for input and output signals
	virtual int calculateDegreesSignals (__VECTOR_INT& argIn, __VECTOR_INT& argOut) const;

	// calculates min and max distances from input to output using connectivity matrix
	// returns Q_OK on success
	int calculateMinMaxDistances (int& minDist, int& maxDist) const;

	virtual int print2Stream (ostream& out) const;
};

};

#endif
