/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#ifndef		__Q_EC_H__
#define		__Q_EC_H__

#include "../../qtools/qtools.h"
#include "../../qtools/qlightxml.h"

namespace ECWorkshop {

// Evolutionary algorithm type enumeration
//	QEAT_GENETIC_ALGORITHM -- means genetic algorithm (with selection crossover and mutation)
//	QEAT_COEVOLUTIONARY_ALGORITHM -- means coevolutionary algorithm (as multiple simultaneous genetic algorithms)
enum QEAType {QEAT_UNKNOWN=0, QEAT_GENETIC_ALGORITHM, QEAT_COEVOLUTIONARY_ALGORITHM, QEAT_SELF_ADAPTIVE_REGULATORY_NETWORK};

// Genetic encoding type enumeration
//	QET_STRING -- string encoding
//	QET_INTEGER -- integer encoding
//	QET_REAL -- real encoding
//	QET_PERMUTATION -- permutation encoding
enum QEncodingType {QET_UNKNOWN=0, QET_STRING, QET_INTEGER, QET_REAL, QET_PERMUTATION, QET_REGULATORY_NETWORK};

// Initialization type enumeration
//	QIT_RANDOM -- random initialization
//	QIT_UNMIXED -- unmixed initialization (!!! for string encoding only !!!)
enum QInitializationType {QIT_UNKNOWN=0, QIT_RANDOM, QIT_UNMIXED};

// Niching operator enumeration
enum QNichingType {QNT_UNKNOWN=0, QNT_SHARING, QNT_CROWDING_DETERMINISTIC, QNT_CROWDING};

// Selection operator enumeration
enum QSelectionType {QST_UNKNOWN=0, QST_NEUTRAL, QST_TOURNAMENT};

// Parent selection operator enumeration
enum QParentSelectorType {QPST_UNKNOWN=0, QPST_RANDOM};

// Population size adaptation operator enumeration
enum QPopulationSizerType {QPSZT_UNKNOWN=0, QPSZT_CONST, QPSZT_SIMPLE, QPSZT_SIMPLE_FIBONACCI, QPSZT_STEADY};

// Crossover operator enumeration
enum QCrossoverType {QXT_UNKNOWN=0, QXT_1POINT, QXT_2POINT, QXT_UNIFORM, QXT_1POINT_PGX, QXT_2POINT_PGX, QXT_BLX, QXT_SBX, QXT_ARITHMETIC, QXT_ORDER};

// Mutation operator enumeration
enum QMutationType {QMT_UNKNOWN=0, QMT_BITFLIP, QMT_RANDOMFLIP, QMT_GAUSS, QMT_RN_BASIC};

// Enumeration for two-side direction
enum QDirection {QD_UNKNOWN=0, QD_UPWARD, QD_DOWNWARD};

// Enumeration for ordering type
enum QOrder {QO_UNKNOWN=0, QO_ASCENDING, QO_DESCENDING};	// QO_UNKNOWN means no order

//
// Stores data which is necessary to run an EA
//
class QProblem;
class QIndividual;
class QEvolutionaryAlgorithm;

struct QEAParameters {
	unsigned seed;			// random numbers generator seed

	QEAType	eaType;

	int populationsNumber;	// number of independent populations for parallel EA (not done yet)

	QEncodingType encoding;	// type of encoding
	bool permutation;		// denotes whether permutation encoding is used. This parameter is to be set automatically
	QOrder order;			// order of permutation sorting (see QOrder enumeration for details)
	int chromosomeLength;	// number of genes
	int arity;				// arity of the symbols (for [string] encoding)
	int geneSize;			// number of bits per gene
	double geneOffset;		// minimal value of the decoded gene
	double genePrecision;	// the minimal step between two neighboring gene values
	int populationSize;		// number of individuals in population
	QInitializationType initializationType;	// type of initialization
	double unitsRatio;		// ratio of units for [integer] initialization
	int generationsNumber;	// number of generations to evolve
	QNichingType nichingType;	// type of niching
	double sharingSigma;	// [sigma] variable value for sharing niching
	QSelectionType selectionType;	// type of selection operator
	int tournamentSize;		// tournament size in tournament selection
	QParentSelectorType parentSelectorType;	// type of parent selection operator
	QPopulationSizerType populationSizerType;
	int minPopSize, maxPopSize;
	int sizingStep;
	int sizingDirection;	// 0 -- stop, 1 -- up, -1 -- down
	QCrossoverType xType;	// type of crossover under use
	QMutationType mutationType;	// type of mutation under use
	int eliteCount;			// number of elite individuals (0 -- no elitism)
	int parentsNumber;		// number of parents that take part in single mating
	int childrenNumber;		// number of children that are generated in result of single mating
	double xRate;			// crossover rate
	double mRate;			// mutation rate
	double xAlpha;			// alpha parameter for BLX and SBX crossovers
	bool removeClones;		// flag for duplicate-chromosomes check
	double errorLevel;		// target value of solutions' error maximum
	QProblem* problem;		// pointer to the problem that is to be solved by EA
	QEvolutionaryAlgorithm* ea;		// pointer to the EA under use

	int inputs;				// number of inputs for ANN and RN
	int outputs;			// number of outputs for ANN and RN

	bool untilFirstHit;

	QEAParameters ();
	~QEAParameters ();

	int print2stream (ostream& out) const;
	int readFromXML (const QLightXMLIO& xml);
};

//
// classes that define type of individuals' fitnesses
//

class QFitness {
protected:

	double _value;
	__VECTOR_DOUBLE _extra;

public:

	QFitness ();
	QFitness (const QFitness& argFitness);
	~QFitness ();

	QFitness& operator= (const QFitness& argFitness);
	QFitness& operator= (const double argD);

	double getValue () const;
	int setValue (const double argValue);
	double getExtra (const int argIndex) const;
	int setExtra (const int argIndex, const double argValue);
	int addExtra (const double argValue);

	int print2Stream (ostream& out) const;
};

};	// namespace

#endif

