/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qintegercrossover.h"

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//		QIntegerCrossover class
//
///////////////////////////////////////////////////////////

QIntegerCrossover::QIntegerCrossover (void) {
}

QIntegerCrossover::QIntegerCrossover (const QEvolutionaryAlgorithm* owner)
: QCrossoverOperator(owner) {
}

QIntegerCrossover::QIntegerCrossover (const QIntegerCrossover& argX)
: QCrossoverOperator(argX) {
}

QIntegerCrossover::~QIntegerCrossover (void) {
}

QIntegerCrossover& QIntegerCrossover::operator= (const QCrossoverOperator& argX) {
	QCrossoverOperator::operator = (argX);

	return (*this);
}

QEncodingType QIntegerCrossover::getEncodingType () const {
	return QET_INTEGER;
}


///////////////////////////////////////////////////////////
//
//		QIntegerCrossover1 class
//
// Notes:
// --------------------------
// 2-point crossover operator for integer encoding
//
///////////////////////////////////////////////////////////

QIntegerCrossover1::QIntegerCrossover1 (void) {
}

QIntegerCrossover1::QIntegerCrossover1 (const QEvolutionaryAlgorithm* owner)
: QIntegerCrossover(owner) {
}

QIntegerCrossover1::QIntegerCrossover1 (const QIntegerCrossover1& argX)
: QIntegerCrossover(argX) {
}

QIntegerCrossover1::~QIntegerCrossover1 (void) {
}

QIntegerCrossover1& QIntegerCrossover1::operator= (const QIntegerCrossover1& argX) {
	this->QIntegerCrossover::operator = (argX);

	return (*this);
}

int QIntegerCrossover1::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* children) {

	int genesNumber = _params->chromosomeLength, geneSize = _params->geneSize, geneMask = (int)splitMask[geneSize];
	int totalBits = genesNumber * geneSize;
	int value1, value2;
	int coPoint;	// break point (somewhere inside the chromosome)
	int left, right, i;
	QIntegerGene *gene1, *gene2, tempGene;

	coPoint = (int)(QRandN() * (totalBits - 1)) + 1;

	for (i=0; coPoint >= geneSize; coPoint -= geneSize, i++) {
		(*children)[0]->setGene (i, (*argParents)[0]->getGene(i));
		(*children)[1]->setGene (i, (*argParents)[1]->getGene(i));
	}

	right = (int)splitMask [coPoint];
	left = geneMask ^ right;

	gene1 = (QIntegerGene*)(*argParents)[0]->getGene (i);
	value1 = gene1->getValue ();
	gene2 = (QIntegerGene*)(*argParents)[1]->getGene (i);
	value2 = gene2->getValue ();

	tempGene.setValue (((value1&left) | (value2&right)));
	(*children)[0]->setGene (i, &tempGene);

	tempGene.setValue (((value2&left) | (value1&right)));
	(*children)[1]->setGene (i, &tempGene);

	i++;
	for (; i < genesNumber; i++) {
		(*children)[0]->setGene (i, (*argParents)[1]->getGene(i));
		(*children)[1]->setGene (i, (*argParents)[0]->getGene(i));
	}

	return Q_OK;
}

QCrossoverOperator* QIntegerCrossover1::clone () const {
	return new QIntegerCrossover1 (*this);
}


///////////////////////////////////////////////////////////
//
//		QIntegerCrossover2 class
//
// Notes:
// --------------------------
// 2-point crossover operator for integer encoding
//
///////////////////////////////////////////////////////////

QIntegerCrossover2::QIntegerCrossover2 (void) {
}

QIntegerCrossover2::QIntegerCrossover2 (const QEvolutionaryAlgorithm* owner)
: QIntegerCrossover(owner) {
}

QIntegerCrossover2::QIntegerCrossover2 (const QIntegerCrossover2& argX)
: QIntegerCrossover(argX) {
}

QIntegerCrossover2::~QIntegerCrossover2 (void) {
}

QIntegerCrossover2& QIntegerCrossover2::operator= (const QIntegerCrossover2& argX) {
	this->QIntegerCrossover::operator = (argX);

	return (*this);
}

int QIntegerCrossover2::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* children) {

	int genesNumber = _params->chromosomeLength, geneSize = _params->geneSize, geneMask = (int)splitMask[geneSize];
	int totalBits = _params->chromosomeLength * _params->geneSize;
	int coPoint1, coPoint2;	// break points (somewhere inside the chromosome)
	int mid, midNot;
	unsigned left, right;
	QIntegerGene *gene1, *gene2, tempGene;
	int value1, value2;
	int i;

	// select crossover points
	coPoint1 = (int)(QRandN() * (totalBits));
	do {
		coPoint2 = (int)(QRandN() * (totalBits));
	}	while (coPoint2 == coPoint1);

	if (coPoint1 > coPoint2) {
		swap (coPoint1, coPoint2);
	}

	for (i=0; coPoint1 >= geneSize; i++, coPoint1-=geneSize, coPoint2 -= geneSize) {
		(*children)[0]->setGene (i, (*argParents)[0]->getGene( i ));
		(*children)[1]->setGene (i, (*argParents)[1]->getGene( i ));
	}

	//cout<<i<<" : "<<coPoint1<<", "<<coPoint2<<"\n";
	//cout.flush ();

	//
	// Exchange of middle parts of genomes
	if (coPoint2 >= geneSize) {	// if xover points are not inside one and the same gene

		right = (int)splitMask[coPoint1];
		left = geneMask ^ right;

		gene1 = (QIntegerGene*)(*argParents)[0]->getGene (i);
		value1 = gene1->getValue ();
		gene2 = (QIntegerGene*)(*argParents)[1]->getGene (i);
		value2 = gene2->getValue ();

		tempGene.setValue (((value1&right) | (value2&left)));
		(*children)[0]->setGene (i, &tempGene);

		tempGene.setValue (((value2&right) | (value1&left)));
		(*children)[1]->setGene (i, &tempGene);

		i++;	// proceed to the next gene
		coPoint2 -= geneSize;

		for (; coPoint2 >= geneSize; i++, coPoint2 -= geneSize) {
			(*children)[0]->setGene (i, (*argParents)[1]->getGene(i));
			(*children)[1]->setGene (i, (*argParents)[0]->getGene(i));
		}

		//cout<<i<<" : "<<coPoint2<<"\n";
		//cout.flush ();

		right = (int)splitMask[coPoint2];
		left = geneMask ^ right;

		gene1 = (QIntegerGene*)(*argParents)[0]->getGene (i);
		value1 = gene1->getValue ();
		gene2 = (QIntegerGene*)(*argParents)[1]->getGene (i);
		value2 = gene2->getValue ();

		tempGene.setValue (((value1&left) | (value2&right)));
		(*children)[0]->setGene (i, &tempGene);

		tempGene.setValue (((value2&left) | (value1&right)));
		(*children)[1]->setGene (i, &tempGene);

		i++;	// proceed to the next gene

	} else {	// coPoint2 < geneSize (xover points are inside one and the same gene)

		mid = (int)(splitMask[coPoint1] ^ splitMask[coPoint2]);
		midNot = geneMask ^ mid;

		gene1 = (QIntegerGene*)(*argParents)[0]->getGene (i);
		value1 = gene1->getValue ();
		gene2 = (QIntegerGene*)(*argParents)[1]->getGene (i);
		value2 = gene2->getValue ();

		tempGene.setValue (((value1&midNot) | (value2&mid)));
		(*children)[0]->setGene (i, &tempGene);

		tempGene.setValue (((value2&midNot) | (value1&mid)));
		(*children)[1]->setGene (i, &tempGene);

		i++;	// proceed to the next gene
	}	// end of if (coPoint2 >= geneSize)
	//cout<<i<<"\n";
	//cout.flush ();

	for (; i<genesNumber; i++) {
		(*children)[0]->setGene (i, (*argParents)[0]->getGene( i ));
		(*children)[1]->setGene (i, (*argParents)[1]->getGene( i ));
	}

	return Q_OK;
}

QCrossoverOperator* QIntegerCrossover2::clone () const {
	return new QIntegerCrossover2 ((QIntegerCrossover2&)(*this));
}

///////////////////////////////////////////////////////////
//
//		QIntegerCrossoverUniform class
//
// Notes:
// --------------------------
// Uniform crossover operator for [integer] encoding
//
///////////////////////////////////////////////////////////

QIntegerCrossoverUniform::QIntegerCrossoverUniform (void) {
}

QIntegerCrossoverUniform::QIntegerCrossoverUniform (const QEvolutionaryAlgorithm* owner)
: QIntegerCrossover(owner) {
}

QIntegerCrossoverUniform::QIntegerCrossoverUniform (const QIntegerCrossoverUniform& argX)
: QIntegerCrossover(argX) {
}

QIntegerCrossoverUniform::~QIntegerCrossoverUniform (void) {
}

QIntegerCrossoverUniform& QIntegerCrossoverUniform::operator= (const QIntegerCrossoverUniform& argX) {
	this->QIntegerCrossover::operator = (argX);

	return (*this);
}

int QIntegerCrossoverUniform::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* children) {
	int i, j, genesNumber = _params->chromosomeLength, geneSize = _params->geneSize, geneMask = (int)splitMask[geneSize];
	int genomeLength = genesNumber * geneSize;
	int value1, value2;
	int mask, mask_not;
	QIntegerGene *gene1, *gene2, tempGene;

	for (i=0; i<genesNumber; i++) {
		gene1 = (QIntegerGene*)(*argParents)[0]->getGene (i);
		value1 = gene1->getValue ();
		gene2 = (QIntegerGene*)(*argParents)[1]->getGene (i);
		value2 = gene2->getValue ();

		mask = 0;
		for (j=0; j<geneSize; j++) {
			if (QRandBool()) {
				mask |= bitMask[j+1];
			}
		}
		mask_not = mask ^ geneMask;

		tempGene.setValue ((value1 & mask) | (value2 & mask_not));
		(*children)[0]->setGene (i, &tempGene);

		tempGene.setValue ((value2 & mask) | (value1 & mask_not));
		(*children)[1]->setGene (i, &tempGene);
	}

	return Q_OK;
}

QCrossoverOperator* QIntegerCrossoverUniform::clone () const {
	return new QIntegerCrossoverUniform ((QIntegerCrossoverUniform&)(*this));
}

///////////////////////////////////////////////////////////
//
//		QIntegerCrossoverPGX1 class
//
// Notes:
// --------------------------
// 1-point per-gene crossover operator for integer encoding
//
///////////////////////////////////////////////////////////

QIntegerCrossoverPGX1::QIntegerCrossoverPGX1 (void) {
}

QIntegerCrossoverPGX1::QIntegerCrossoverPGX1 (const QEvolutionaryAlgorithm* owner)
: QIntegerCrossover(owner) {
}

QIntegerCrossoverPGX1::QIntegerCrossoverPGX1 (const QIntegerCrossoverPGX1& argX)
: QIntegerCrossover(argX) {
}

QIntegerCrossoverPGX1::~QIntegerCrossoverPGX1 (void) {
}

QIntegerCrossoverPGX1& QIntegerCrossoverPGX1::operator= (const QIntegerCrossoverPGX1& argX) {
	this->QIntegerCrossover::operator = (argX);

	return (*this);
}

int QIntegerCrossoverPGX1::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* children) {

	int genesNumber = _params->chromosomeLength, geneSize = _params->geneSize, geneMask = (int)splitMask[geneSize];
	int totalBits = _params->chromosomeLength * _params->geneSize;
	unsigned left, right;
	QIntegerGene *gene1, *gene2, tempGene;
	int value1, value2;
	int i;

	//_fout<<"\n\tParents:";
	//_fout<<"\n1:\t";
	//(*argParents)[0]->print2Stream (_fout);
	//_fout<<"\n2:\t";
	//(*argParents)[1]->print2Stream (_fout);

	//_fout<<"\nmasks:\n"<<hex;

	for (i=0; i<genesNumber; i++) {
		// select crossover point
		right = (int)splitMask [(int)((double)(geneSize) * QRandN())];
		left = geneMask ^ right;

		//_fout<<left<<" : "<<right<<"\n";

		gene1 = (QIntegerGene*)((*argParents)[0]->getGene (i));
		value1 = gene1->getValue ();
		gene2 = (QIntegerGene*)((*argParents)[1]->getGene (i));
		value2 = gene2->getValue ();

		tempGene.setValue (((value1&left) | (value2&right)));
		(*children)[0]->setGene (i, &tempGene);

		tempGene.setValue (((value2&left) | (value1&right)));
		(*children)[1]->setGene (i, &tempGene);

	}
	//_fout<<dec;
	//_fout<<"\tChildren:";
	//_fout<<"\n1:\t";
	//(*children)[0]->print2Stream (_fout);
	//_fout<<"\n2:\t";
	//(*children)[1]->print2Stream (_fout);
	//_fout<<"\n";

	return Q_OK;
}

QCrossoverOperator* QIntegerCrossoverPGX1::clone () const {
	return new QIntegerCrossoverPGX1 ((QIntegerCrossoverPGX1&)(*this));
}


///////////////////////////////////////////////////////////
//
//		QIntegerCrossoverPGX2 class
//
// Notes:
// --------------------------
// 2-point per-gene crossover operator for integer encoding
//
///////////////////////////////////////////////////////////

QIntegerCrossoverPGX2::QIntegerCrossoverPGX2 (void) {
}

QIntegerCrossoverPGX2::QIntegerCrossoverPGX2 (const QEvolutionaryAlgorithm* owner)
: QIntegerCrossover(owner) {
}

QIntegerCrossoverPGX2::QIntegerCrossoverPGX2 (const QIntegerCrossoverPGX2& argX)
: QIntegerCrossover(argX) {
}

QIntegerCrossoverPGX2::~QIntegerCrossoverPGX2 (void) {
}

QIntegerCrossoverPGX2& QIntegerCrossoverPGX2::operator= (const QIntegerCrossoverPGX2& argX) {
	this->QIntegerCrossover::operator = (argX);

	return (*this);
}

int QIntegerCrossoverPGX2::cross (const __VECTOR_CONST_INDIVIDUAL_PTR* argParents, __VECTOR_INDIVIDUAL_PTR* children) {

	int genesNumber = _params->chromosomeLength, geneSize = _params->geneSize, geneMask = (int)splitMask[geneSize];
	int totalBits = _params->chromosomeLength * _params->geneSize;
	int mid, midNot;
	unsigned left, right;
	QIntegerGene *gene1, *gene2, tempGene;
	int value1, value2;
	int i;

	//_fout<<"\n\tParents:";
	//_fout<<"\n1:\t";
	//(*argParents)[0]->print2Stream (_fout);
	//_fout<<"\n2:\t";
	//(*argParents)[1]->print2Stream (_fout);
	//_fout<<"\nmasks:\n"<<hex;

	for (i=0; i<genesNumber; i++) {

		// select crossover points
		right = (int)(QRandN() * (geneSize));
		do {
			left = (int)(QRandN() * (geneSize));
		}	while (left == right);

		right = (int)splitMask [right];
		left = (int)splitMask [left];
		mid = left ^ right;
		midNot = geneMask ^ mid;

		//_fout<<mid<<" : "<<midNot<<"\t"<<left<<" : "<<right<<"\n";

		gene1 = (QIntegerGene*)((*argParents)[0]->getGene (i));
		value1 = gene1->getValue ();
		gene2 = (QIntegerGene*)((*argParents)[1]->getGene (i));
		value2 = gene2->getValue ();

		tempGene.setValue (((value1 & midNot) | (value2 & mid)));
		(*children)[0]->setGene (i, &tempGene);

		tempGene.setValue (((value2 & midNot) | (value1 & mid)));
		(*children)[1]->setGene (i, &tempGene);

	}
	//_fout<<dec;
	//_fout<<"\tChildren:";
	//_fout<<"\n1:\t";
	//(*children)[0]->print2Stream (_fout);
	//_fout<<"\n2:\t";
	//(*children)[1]->print2Stream (_fout);
	//_fout<<"\n";

	return Q_OK;
}

QCrossoverOperator* QIntegerCrossoverPGX2::clone () const {
	return new QIntegerCrossoverPGX2 ((QIntegerCrossoverPGX2&)(*this));
}

