/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qrnmutation.h"

#include "../../encodings/include.h"

#include <algorithm>
#include <math.h>

using namespace ECWorkshop;


///////////////////////////////////////////////////////////
//
//		QRNMutation class
//
// Note:
// ------------
// Base class for mutation operators for [real] encoding
//
///////////////////////////////////////////////////////////


QRNMutation::QRNMutation (void) {
}

QRNMutation::QRNMutation (const QEvolutionaryAlgorithm* owner)
: QMutationOperator(owner) {
}

QRNMutation::QRNMutation (const QRNMutation& argMut)
: QMutationOperator(argMut) {
}

QRNMutation::~QRNMutation (void) {
}

QRNMutation& QRNMutation::operator= (const QRNMutation& argMut) {
	this->QMutationOperator::operator = (argMut);
	return (*this);
}

QEncodingType QRNMutation::getEncodingType () const {
	return QET_REGULATORY_NETWORK;
}


///////////////////////////////////////////////////////////
//
//		QRNMutationBasic class
//
// Note:
// ------------
// Gaussian mutation operator for [real] encoding
//
///////////////////////////////////////////////////////////

QRNMutationBasic::QRNMutationBasic (void) {
}

QRNMutationBasic::QRNMutationBasic (const QEvolutionaryAlgorithm* owner)
: QRNMutation(owner) {
}

QRNMutationBasic::QRNMutationBasic (const QRNMutationBasic& argMut)
: QRNMutation(argMut) {
}

QRNMutationBasic::~QRNMutationBasic (void) {
}

QRNMutationBasic& QRNMutationBasic::operator= (const QRNMutationBasic& argMut) {
	this->QRNMutation::operator= (argMut);
	return (*this);
}

int QRNMutationBasic::operate () {
	int i, j, size = _mutant->getSize(), mType;
	double mRate, tempSum, temp;
	QRNIndividual* tempInd = (QRNIndividual*)(_mutant);

	// define mRate value using averaged absolute values of
	// mutation parameters
	mRate = tempSum = 0.0;
	//cout<<"\n======================================================\nMutation parameters:\t";

	// temporal (?) decision to slow down the RN's growth
	//tempInd->setMutationParam(SPLIT_NODE, tempInd->getMutationParam(SPLIT_NODE) / 2.0);
	for (i=0; i<MUTATION_PARAMETERS; i++) {
		mRate += fabs( tempInd->getMutationParam(i) );
		tempSum += fabs(tempInd->getMutationParam(i));
		//cout<<tempInd->getMutationParam(i)<<" : ";
	}
	//cout<<"\n";
	mRate /= (double)MUTATION_PARAMETERS;
	mRate *= _params->mRate;

	// For debug purposes *only*
	//mRate = 1.0;
	// For debug purposes *only*

	//cout<<"mRate = "<<mRate<<"\n";

	/////////////////////////////////////////////////////////////////
	// RN mutation procedure
	// 0. Pick node
	// 1. For each node define whether the mutation occurs (via mRate variable)
	// 2. Define at random what kind of mutation (1 of 9 possible ones) takes place
	// 3. Apply selected type of mutation
	// 4. Proceed to the next node

	// cycle through all nodes (step 0)
//	cout<<"size = "<<size<<" : tempSum = "<<tempSum<<"\n";
	for (i=0; i<size; i++) {
		if (mRate > QRandN()) {	// viva la mutation! (step 1)
			// step 2
			// choose the mutation type via a roulette-like procedure
			temp = tempSum * QRandN();
//			cout<<"temp = "<<temp<<"\n";
			mType = 0;
			for (j=0; j<MUTATION_PARAMETERS; j++) {
				temp -= fabs(tempInd->getMutationParam(j));
				if (temp <= 0) {
					mType = j;
					break;
				}
			}

			// decay [tempSum]
			//tempSum -= mRate*QRandN();
			tempSum -= tempSum*QRandN();

//			cout<<"mType = "<<mType<<"\n";
			//mType = NODE_TYPE_CHANGE;

			switch (mType) {
				case ADD_DELETE_NODE:	//	0
				//	//cout<<"mutateAddDeleteNode ("<<tempInd->getMutationParam(ADD_DELETE_NODE)<<")...";
				//	//cout.flush ();
					mutateAddDeleteNode (tempInd->getMutationParam(ADD_DELETE_NODE));
				//	//cout<<"done\n";
				//	//cout.flush ();
					break;
				case ADD_DELETE_INPUT:	//	1
					//cout<<"mutateAddDeleteInput...";
					//cout.flush ();
					mutateAddDeleteInput (tempInd->getMutationParam(ADD_DELETE_INPUT));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case ADD_DELETE_OUTPUT:	//	2
					//cout<<"mutateAddDeleteOutput...";
					//cout.flush ();
					mutateAddDeleteOutput (tempInd->getMutationParam(ADD_DELETE_OUTPUT));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case NODE_TYPE_CHANGE:	//	3
					//cout<<"mutateChangeNode...";
					//cout.flush ();
					mutateChangeNode (tempInd->getMutationParam(NODE_TYPE_CHANGE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case INPUT_TYPE_CHANGE:	//	4
					//cout<<"mutateChangeInput...";
					//cout.flush ();
					mutateChangeInput (tempInd->getMutationParam(INPUT_TYPE_CHANGE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case OUTPUT_TYPE_CHANGE:	//	5
					//cout<<"mutateChangeOutput...";
					//cout.flush ();
					mutateChangeOutput (tempInd->getMutationParam(OUTPUT_TYPE_CHANGE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case INPUT_WEIGHT_CHANGE:	//	6
					//cout<<"mutateChangeInputWeight...";
					//cout.flush ();
					mutateChangeInputWeight (tempInd->getMutationParam(INPUT_WEIGHT_CHANGE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case OUTPUT_WEIGHT_CHANGE:	//	7
					//cout<<"mutateChangeOutputWeight...";
					//cout.flush ();
					mutateChangeOutputWeight (tempInd->getMutationParam(OUTPUT_WEIGHT_CHANGE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case TOGGLE_NODE_ON_OFF:	//	8
					//cout<<"mutateOnOffNode...";
					//cout.flush ();
					mutateOnOffNode (tempInd->getMutationParam(TOGGLE_NODE_ON_OFF));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case SPLIT_NODE:			//	9
					//cout<<"mutateSplitNode...";
					//cout.flush ();
					mutateSplitNode (tempInd->getMutationParam(SPLIT_NODE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				case NODE_BIAS_CHANGE:			//	9
					//cout<<"mutateSplitNode...";
					//cout.flush ();
					mutateChangeNodeBias (tempInd->getMutationParam(NODE_BIAS_CHANGE));
					//cout<<"done\n";
					//cout.flush ();
					break;
				default:
					cout<<"What the heck?\n";
					cout.flush ();
					break;
			}	// switch (mType)
			//_mutant->print2Stream (cout);
			tempInd->writeHistory (mType);
		}	// if (mRate > QRandN())

		// write history event
		if (tempSum < 0) {	// no more mutations allowed
			break;
		}

	}	// for (i=0; i<size; i++)

	////////////////////////////////////////////////
	// Mutate mutation parameters
	__VECTOR_DOUBLE hist;

	tempSum = 0.0;
	for (i=0; i<MUTATION_PARAMETERS; ++i) {
		tempSum += tempInd->readHistory (i);
	}
	hist.assign (MUTATION_PARAMETERS, 1.0/MUTATION_PARAMETERS);
	if (tempSum > 0) {	// if the mutation took place in the past
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			hist[i] += (double)tempInd->readHistory(i) / tempSum;
		}
	} else {	// do mutation of mutation parameters another way
		for (i=0; i<MUTATION_PARAMETERS; ++i) {
			hist[i] = 1.0/MUTATION_PARAMETERS;
		}
	}
	//mRate /= _params->mRate;
	//if (mRate > 1.0) {
	//	mRate = 1.0/mRate;
	//}
	for (i=0; i<MUTATION_PARAMETERS; i++) {
		//temp = tempInd->getMutationParam(i) + QRandGauss() * hist[i];
		//tempInd->setMutationParam (i, temp);
		tempInd->setMutationParam (i, QRandGauss() * hist[i]);
	}

	return Q_OK;
}

QMutationOperator* QRNMutationBasic::clone () const {
	return new QRNMutationBasic ((QRNMutationBasic&)(*this));
}

//
// *** protected methods ***
//

int QRNMutationBasic::mutateChangeNodeBias (const double argMParam) {
	int i, maxDgr;
	QRNGene* tempGene;
	QRNIndividual *tempInd = (QRNIndividual*)_mutant;
	__VECTOR_INT dgr;

	// select node
	tempInd->calculateDegrees(dgr);
	if (argMParam <= 0) {
		// invert [dgr]
		maxDgr = (*max_element(dgr.begin(), dgr.end()));
		for (i=0; i<(int)dgr.size(); i++) {
			dgr[i] = maxDgr - dgr[i];
		}
	}
	i = roulette (dgr);
	tempGene = (QRNGene*)(tempInd->getGene (i));
	tempGene->setBias ( tempGene->getBias () + fabs(argMParam) * QRandGauss() );

	return Q_OK;
}

int QRNMutationBasic::mutateAddDeleteNode (const double argMParam) {
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	__VECTOR_INT dgr, dgrSigIn, dgrSigOut;	// arrays to store node degrees
									// and signals (input/output) degrees
	int i, tempSum, maxDgr, size, delNode;
	double temp;

	if (argMParam > 0) {	// add node
		//cout<<"Add node...\n";
		QRNGene tempGene;
		int sumIn, sumOut;
		__VECTOR_SIGNAL ins, outs;
		QSignal tempSignal;

		tempInd->calculateDegreesSignals (dgrSigIn, dgrSigOut);
		// if arrays for input and output signals are of different sizes
		// then enlarge the smaller one
		if (dgrSigIn.size() > dgrSigOut.size()) {
			// enlarge [dgrSigOut]
			do {
				dgrSigOut.push_back (0);
			} while (dgrSigIn.size() > dgrSigOut.size());

		} else if (dgrSigIn.size() < dgrSigOut.size()) {
			// enlarge [dgrSigIn]
			do {
				dgrSigIn.push_back (0);
			} while (dgrSigIn.size() < dgrSigOut.size());
		}	// otherwise do nothing
		//dgrSigIn.push_back (0);		// add a new signal id
		//dgrSigOut.push_back (0);	// add a new signal id

		/* TODO (Qai#1#): How to choose number of input/output signals? */

		// calculate sums of input/output signals degrees
		//
		//	******************************************
		//
		// Important! RN output signal can not be send to an input
		// to any node as well as RN input signals can not be 
		// resulted as some node's output
		//
		// This is a temporal case and I'm going to reconsider
		// it in the future. Maybe not...
		//
		//	[Y. Tsoy]
		//
		//	******************************************
		//
		size  = (int)dgrSigOut.size ();
		sumIn = sumOut = 0;
		for (i=0; i<size; i++) {
			// sum only those degrees which signals can be used as output
			if (i == 0 || i > _params->inputs) {
				dgrSigIn[i] += INITIAL_VALUE;
				sumIn += dgrSigIn[i];
			} else {
				dgrSigIn[i] = 0;
			}

			// sum only those degrees which signals can be used as input
			if (i <= _params->inputs || i > (_params->inputs + _params->outputs)) {
				dgrSigOut[i] += INITIAL_VALUE;
				sumOut += dgrSigOut[i];
			} else {
				dgrSigOut[i] = 0;
			}

			//if (i > 0) {
			//	dgrSigOut[i] += INITIAL_VALUE;
			//	sumOut += dgrSigOut[i];
			//} else {
			//	dgrSigOut[i] = 0;
			//}

		}	// for (i=0; i<size; i++)

		for (i=0; i<size; i++) {
			// temp is propotional to the degree of the i-th signal
			temp = (double)dgrSigIn[i] / (double)sumIn;
			// if temp is greater than random value than add the i-th input
			if (temp > QRandN()) {
				tempSignal._id = i;
				outs.push_back (tempSignal);
			}

			// if temp is greater than random value than add the i-th output
			temp = (double)dgrSigOut[i] / (double)sumOut;
			if (temp > QRandN()) {
				tempSignal._id = i;
				ins.push_back (tempSignal);
			}
		}	// for (i=0; i<size; i++)
		// if input/output signals arrays are empty
		// then add an arbitrary input/output signal
		if (ins.empty ()) {
			// TODO: new [id] should be selected according to the actual number of
			// input signals (not the initial one)
			tempSignal._id = (int)(((double)dgrSigOut.size()) * QRandN() + 0.5);
			if (tempSignal._id > _params->inputs && tempSignal._id <= (_params->inputs + _params->outputs)) {
				tempSignal._id = 0;
			}
			ins.push_back (tempSignal);
		}
		if (outs.empty ()) {
			// TODO: new [id] should be selected according to the actual number of
			// ouput signals (not the initial one)
			tempSignal._id = (int)((double)dgrSigIn.size() * QRandN() + 0.5);
			if (tempSignal._id <= _params->inputs) {
				tempSignal._id = 0;
			}
			outs.push_back (tempSignal);
		}
		// set input and output signals
		tempGene.setInputs (ins);
		tempGene.setOutputs (outs);

		// set input weights
		size = (int)ins.size ();
		for (i=0; i<size; i++) {
			tempGene.setInputWeight (i, 0.2*QRandN() - 0.1);
		}

		// set output weights
		size = (int)outs.size ();
		for (i=0; i<size; i++) {
			tempGene.setOutputWeight (i, 0.2*QRandN() - 0.1);
		}

		// set bias
		tempGene.setBias (0.2*QRandN() - 0.1);

		/* TODO (Qai#1#): How to choose new node type? */
		// meanwhile the new node type is QRNT_SIGMOID (default)
		//tempGene.setType (QRNT_SIGMOID);
		i = (int)((QRNT_SIZE-1.0) * QRandN() + 1);
		tempGene.setType ((QRNNodeType)i);
		tempGene.setParamA (10.0*QRandGauss());

		tempInd->addGene (&tempGene);
		//tempGene.print2Stream (cout);

	} else {	// delete node according to its degree
				// (the less the degree, the more the probability that
				// the node would be deleted)
		//cout<<"Delete node...\n";
		size = tempInd->getSize();
		if (size <= 1) {
			return Q_FAILURE;
		}

		// calculate degrees
		tempInd->calculateDegrees (dgr);
		maxDgr = (*max_element (dgr.begin(), dgr.end()));

		tempSum = 0;
		for (i=0; i<size; i++) {
			// "invert" values of degrees for proper selection of node to delete
			dgr[i] = maxDgr - dgr[i] + INITIAL_VALUE;
			tempSum += dgr[i];
		}

		// select node to delete
		temp = tempSum * QRandN();
		for (i=0; i<size; i++) {
			temp -= dgr[i];
			if (temp < 0) {
				delNode = i;
				break;
			}
		}

		tempInd->deleteGene (delNode);
	}	// if (argMParam > 0)

	return Q_OK;
}

// The gene undergoing mutation should be defined according to its degree
// inside the method
int QRNMutationBasic::mutateAddDeleteInput (const double argMParam) {
	int i, size, tempSum, temp;
	QRNGene* tempGene;
	QRNIndividual *tempInd = (QRNIndividual*)_mutant;
	__VECTOR_INT dgr;

	tempInd->calculateDegrees(dgr);
	i = roulette (dgr);

	tempGene = (QRNGene*)(tempInd->getGene (i));
	
	__VECTOR_INT dgrIn, dgrOut;
	int maxDgr;
	tempInd->calculateDegreesSignals (dgrIn, dgrOut);
	QSignal tempSignal;

	// select whether input signal would be added or deleted
	if (argMParam > 0) {	// add input signal
							// the type of newly added signal is defined
							// by the degree of already existing ones

		// if array for input signals is larger
		// then enlarge the output signals array
		if (dgrIn.size() > dgrOut.size()) {
			// enlarge [dgrSigOut]
			do {
				dgrOut.push_back (0);
			} while (dgrIn.size() > dgrOut.size());
		}
		//dgrOut.push_back (0);	// add a new signal id

		size = (int)dgrOut.size();
		tempSum = 0;
		for (i=0; i<size; i++) {
			// RN output signal can not be selected as input
			// See [mutateAddDeleteNode] for details
			if (i <= _params->inputs || i > (_params->inputs + _params->outputs)) {
				dgrOut[i] += INITIAL_VALUE;
				tempSum += dgrOut[i];
//				cout<<dgrOut[i]<<"\n";
			} else {
				dgrOut[i] = 0;
			}

		}
		// define type of a new signal
		temp = (int)(tempSum * QRandN() + 0.5);
		for (i=0; i<size; i++) {
			temp -= dgrOut[i];
			if (temp <= 0) {
				// the i-th signal is selected
				break;
			}
		}
		tempSignal._id = i;
		tempGene->addInputSignal (&tempSignal, (0.2*QRandN()-0.1));

	} else {	// delete input signal using signals degree
				// the more the degree the less the probability that the
				// signal is to be deleted

		if (tempGene->getInputsNumber() <= 1) {
			return Q_FAILURE;
		}

		// "invert" input signals' degrees
		size = (int)dgrIn.size ();
		maxDgr = (*max_element (dgrOut.begin(), dgrOut.end()));
		tempSum = 0;
		for (i=0; i<size; i++) {
			if (i > 0 && tempGene->hasInput(i)) {
				dgrIn[i] = maxDgr - dgrIn[i] + INITIAL_VALUE;
				tempSum += dgrIn[i];
			} else {
				dgrIn[i] = 0;
			}
		}

		temp = (int)(tempSum * QRandN() + 0.5);
//		cout<<"Temp sum = "<<tempSum<<"\n";
//		cout<<"temp = "<<temp<<"\n";
		for (i=0; i<size; i++) {
			if (tempGene->hasInput(i)) {
				temp -= dgrIn[i];
				if (temp <= 0) {
					break;
				}
			}
		}
		tempGene->deleteInputSignalByID (i);

	}	// if (argMParam > 0)

	return Q_OK;
}

int QRNMutationBasic::mutateAddDeleteOutput (const double argMParam) {
	int i, size, tempSum, temp;
	QRNGene* tempGene;
	QRNIndividual *tempInd = (QRNIndividual*)_mutant;
	__VECTOR_INT dgr;

	tempInd->calculateDegrees(dgr);
	i = roulette (dgr);

	tempGene = (QRNGene*)(tempInd->getGene (i));

	__VECTOR_INT dgrIn, dgrOut;
	int maxDgr;
	tempInd->calculateDegreesSignals (dgrIn, dgrOut);
	QSignal tempSignal;

	// select whether output signal would be added or deleted
	if (argMParam > 0) {	// add output signal
							// the type of newly added signal is defined
							// by the degree of already existing ones

		// if array for output signals is large than
		// then enlarge the input signals array
		if (dgrOut.size() > dgrIn.size()) {
			// enlarge [dgrIn]
			do {
				dgrIn.push_back (0);
			} while (dgrOut.size() > dgrIn.size());
		}
		//dgrIn.push_back (0);	// add a new signal id

		size = (int)dgrIn.size();
		tempSum = 0;
		for (i=0; i<size; i++) {
			// RN output signal can not be selected as input
			// See [mutateAddDeleteNode] for details
			if (i == 0 || i > _params->inputs) {
				dgrIn[i] += INITIAL_VALUE;
				tempSum += dgrIn[i];
//				cout<<dgrIn[i]<<"\n";
			} else {
				dgrIn[i] = 0;
			}
		}
		// define type of a new signal
		temp = (int)(tempSum * QRandN() + 0.5);
		for (i=0; i<size; i++) {
			temp -= dgrIn[i];
			if (temp <= 0) {
				// the i-th signal is selected
				break;
			}
		}
		tempSignal._id = i;
//		tempGene->addOutputSignal (&tempSignal, (0.2*QRandN()-0.1));
		tempGene->addOutputSignal (&tempSignal, 1);

	} else {	// delete output signal using signals degree
				// the more the degree the less the probability that the
				// signal is to be deleted

		if (tempGene->getOutputsNumber() <= 1) {
			return Q_FAILURE;
		}

		// "invert" output signals' degrees
		size = (int)dgrOut.size ();
		maxDgr = (*max_element (dgrOut.begin(), dgrOut.end()));
		tempSum = 0;
		for (i=0; i<size; i++) {
			if (i > 0 && tempGene->hasOutput(i)) {
				dgrOut[i] = maxDgr - dgrOut[i] + INITIAL_VALUE;
				tempSum += dgrOut[i];
			} else {
				dgrOut[i] = 0;
			}
		}

		temp = (int)(tempSum * QRandN() + 0.5);
//		cout<<"Temp sum = "<<tempSum<<"\n";
//		cout<<"temp = "<<temp<<"\n";
		for (i=0; i<size; i++) {
			if (tempGene->hasOutput(i)) {
				temp -= dgrOut[i];
				if (temp <= 0) {
					break;
				}
			}
		}
		tempGene->deleteOutputSignalByID (i);

	}	// if (argMParam > 0)

	return Q_OK;
}

// change randomly node type
int QRNMutationBasic::mutateChangeNode (const double argMParam) {
	__VECTOR_INT dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	QRNNodeType type;
	int i, maxDgr;

	// select random node using roulette wheel
	tempInd->calculateDegrees (dgr);
	if (argMParam <= 0) {
		// invert [dgr]
		maxDgr = (*max_element(dgr.begin(), dgr.end()));
		for (i=0; i<(int)dgr.size(); i++) {
			dgr[i] = maxDgr - dgr[i];
		}
	}
	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	// select at random new node type
	do {
		type = (QRNNodeType)((int)(((double)QRNT_SIZE - 1.0) * QRandN () + 1.0));
	} while (tempGene->getType() == type);
	tempGene->setType (type);
	//tempGene->setType (QRNT_SIN);

	//if (tempGene->getType() != QRNT_SIN) {
	//	cout<<"nooooo!\n";
	//	exit (0);
	//}

	//if (type == QRNT_SIGMOID) {
	//	cout<<"z";
	//}

	return Q_OK;
}

int QRNMutationBasic::mutateChangeInput (const double argMParam) {
	__VECTOR_INT dgrIn, dgrOut, dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	int i, idx, maxDgr, newID;

	//
	// select random gene
	tempInd->calculateDegrees (dgr);
	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	if (tempGene->getInputsNumber() == 0) {
		return Q_FAILURE;
	}

	//
	// select random input using roulette wheel
	tempInd->calculateDegreesSignals (dgrIn, dgrOut);
	// if dgrIn.size > dgrOut.size then
	// resize dgrOut array
	while (dgrIn.size() > dgrOut.size()) {
		dgrOut.push_back (0);
	}

	if (dgrOut.size() <= 1) {
		// there is nothing to select from
		return Q_FAILURE;
	}

	//dgrOut.push_back (0);	// add a new signal id
	// select signal type to change
	if (argMParam <= 0) {
		// invert [dgrOut]
		maxDgr = (*max_element(dgrOut.begin(), dgrOut.end()));
		for (i=0; i<(int)dgrOut.size(); i++) {
			dgrOut[i] = maxDgr - dgrOut[i];
		}
	}
	// select only that input signal which is already present at the node
	do {
		idx = roulette (dgrOut, INITIAL_VALUE);
	} while (!tempGene->hasInput(idx));
	// select new signal ID
	dgrOut[idx] = -INITIAL_VALUE;	// to avoid selection of the same signal id again
	for (i=_params->inputs+1; i<=(_params->inputs + _params->outputs); i++) {
		dgrOut[i] = -INITIAL_VALUE;	// to null the element during roulette selection
	}
	int count=0;
	do {
		newID = roulette (dgrOut, INITIAL_VALUE);
		count++;
		if (count > 100) {
			for (i=0; i<(int)dgrOut.size(); i++) {
				cout<<" "<<dgrOut[i];
			}
			cout<<"\n";
			cout.flush ();
			exit (0);
		}
	} while (newID == idx);
	tempGene->changeInputID (idx, newID);

	return Q_OK;
}

int QRNMutationBasic::mutateChangeOutput (const double argMParam) {
	__VECTOR_INT dgrIn, dgrOut, dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	int i, idx, maxDgr, newID;

	//
	// select random gene
	tempInd->calculateDegrees (dgr);
	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	if (tempGene->getOutputsNumber() == 0) {
		return Q_FAILURE;
	}

	// calculate signals degrees and make [dgrIn]
	// vector to have the same size as [dgrOut]
	tempInd->calculateDegreesSignals (dgrIn, dgrOut);
	if (dgrIn.size() < dgrOut.size()) {
		do {
			dgrIn.push_back (0);
		} while (dgrIn.size() < dgrOut.size());
	}

	if (dgrIn.size() <= 1) {
		// there is nothing to select from
		return Q_FAILURE;
	}

	//
	// select random output using roulette wheel
	//
	//dgrIn.push_back (0);	// add a new signal id
	if (argMParam <= 0) {
		// invert [dgrIn]
		maxDgr = (*max_element(dgrIn.begin(), dgrIn.end()));
		for (i=0; i<(int)dgrIn.size(); i++) {
			dgrIn[i] = maxDgr - dgrIn[i];
		}
	}
	// select only that output signal which is already present at the node
	do {
		idx = roulette (dgrIn, INITIAL_VALUE);
	} while (!tempGene->hasOutput(idx));
	// select new signal ID
	dgrIn[idx] = -INITIAL_VALUE;	// to avoid selection of the same signal id again
	for (i=1; i<= _params->inputs; i++) {
		dgrIn[i] = -INITIAL_VALUE;	// to avoid selection of input signals as an output
	}
	do {
		newID = roulette (dgrIn, INITIAL_VALUE);
	} while (newID == idx);
	tempGene->changeOutputID (idx, newID);

	return Q_OK;
}

int QRNMutationBasic::mutateChangeInputWeight (const double argMParam) {
	__VECTOR_INT dgrIn, dgrOut, dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	int i, idx, maxDgr, size;

	//
	// select random gene
	tempInd->calculateDegrees (dgr);
	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	if (tempGene->getInputsNumber() == 0) {
		return Q_FAILURE;
	}

	tempInd->calculateDegreesSignals (dgrIn, dgrOut);

	// TODO:
	// select input to change

	// define degrees for each input
	__VECTOR_INT inputDgr;
	inputDgr.resize (tempGene->getInputsNumber());
	size = (int)inputDgr.size();
	for (i=0; i<size; i++) {
		inputDgr[i] = dgrIn [ tempGene->getInput(i)->_id ];
	}

	if (argMParam <= 0) {
		// invert [inputDgr]
		maxDgr = (*max_element(inputDgr.begin(), inputDgr.end()));
		size = (int)inputDgr.size();
		for (i=0; i<size; i++) {
			inputDgr[i] = maxDgr - inputDgr[i];
		}
	}
	// select input using roulette
	idx = roulette (inputDgr, INITIAL_VALUE);

	tempGene->setInputWeight(idx, tempGene->getInputWeight(idx) + 0.2*QRandN() - 0.1);

	return Q_OK;
}

int QRNMutationBasic::mutateChangeOutputWeight (const double argMParam) {
	__VECTOR_INT dgrIn, dgrOut, dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	int i, idx, maxDgr, size;

	//
	// select random gene
	tempInd->calculateDegrees (dgr);
	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	if (tempGene->getOutputsNumber() == 0) {
		return Q_FAILURE;
	}

	tempInd->calculateDegreesSignals (dgrIn, dgrOut);

	// TODO:
	// select output to change

	// define degrees for each output
	__VECTOR_INT outputDgr;
	outputDgr.resize (tempGene->getOutputsNumber());
	size = (int)outputDgr.size();
	for (i=0; i<size; i++) {
		outputDgr[i] = dgrOut [ tempGene->getOutput(i)->_id ];
	}

	if (argMParam <= 0) {
		// invert [inputDgr]
		maxDgr = (*max_element(outputDgr.begin(), outputDgr.end()));
		size = (int)outputDgr.size();
		for (i=0; i<size; i++) {
			outputDgr[i] = maxDgr - outputDgr[i];
		}
	}
	// select input using roulette
	idx = roulette (outputDgr, INITIAL_VALUE);

//	cout<<"idx = "<<idx<<" ()"
	tempGene->setOutputWeight(idx, tempGene->getOutputWeight(idx) + 0.2*QRandN() - 0.1);

	return Q_OK;
}

int QRNMutationBasic::mutateOnOffNode (const double argMParam) {
	__VECTOR_INT dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	int i, maxDgr, size;

	//
	// select random gene
	tempInd->calculateDegrees (dgr);

	if (argMParam <= 0) {
		// invert [dgr]
		maxDgr = (*max_element(dgr.begin(), dgr.end()));
		size = (int)dgr.size();
		for (i=0; i<size; i++) {
			dgr[i] = maxDgr - dgr[i];
		}
	}

	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	// toggle node's state
	tempGene->enable (!tempGene->isEnabled());

	return Q_OK;
}

int QRNMutationBasic::mutateSplitNode (const double argMParam) {
	__VECTOR_INT dgr;
	QRNIndividual* tempInd = (QRNIndividual*)_mutant;
	QRNGene* tempGene;
	QRNGene newGene;
	int i, maxDgr, size;

	//
	// select random gene
	tempInd->calculateDegrees (dgr);

	if (argMParam <= 0) {
		// invert [dgr]
		maxDgr = (*max_element(dgr.begin(), dgr.end()));
		size = (int)dgr.size();
		for (i=0; i<size; i++) {
			dgr[i] = maxDgr - dgr[i];
		}
	}

	i = roulette (dgr, INITIAL_VALUE);
	tempGene = (QRNGene*)tempInd->getGene (i);

	// the selected gene is splitted into two such a way that
	// the 1st gene gets all the inputs and the 2nd one gets all the outputs
	// while these genes are connected via single signal (with a new id!)
	newGene.assign (tempGene);

	QSignal tempSignal;
	tempSignal._id = max (tempInd->getMaxInputID (), tempInd->getMaxOutputID ()) + 1;

	if (tempSignal._id <= _params->inputs + _params->outputs) {
		tempSignal._id = _params->inputs + _params->outputs + 1;
	}

	// set new output for the 1st gene
	tempGene->setOutputsNumber (1);
	tempGene->setOutput (0, tempSignal);
	tempGene->setOutputWeight (0, 1.0);

	// set new input for the 2nd gene
	newGene.setInputsNumber (1);
	newGene.setInput (0, tempSignal);
	newGene.setInputWeight (0, 0.2*QRandN()-0.1);

	tempInd->addGene (&newGene);

	return 0;
}
