/*****************************************

ECWorkshop -- evolutionary computation
toolkit library

Web: http://qai.narod.ru/ecw/
Mailto: qai@mail.ru

Copyright (C) 2007 by Yury Tsoy


This program is free software; you can
redistribute it and/or modify it under the
terms of the GNU General Public License as
published by the Free Software Foundation;
either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that
it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU General Public License
for more details.

You should have received a copy of the GNU
General Public License along with this
program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA  02111-1307  USA

*/

#include "qeafactory.h"

#include "../eas/include.h"

#include "initialization/qintegerinitializer.h"
#include "initialization/qstringinitializer.h"
#include "initialization/qrcinitializer.h"

#include "niching/qfitnesssharing.h"
#include "niching/qcrowding.h"

#include "selection/qneutralselection.h"
#include "selection/qtournamentselection.h"

#include "parentselector/qrandomparentselector.h"

#include "populationsizer/qconstpopulationsizer.h"
#include "populationsizer/qsteadypopulationsizer.h"
#include "populationsizer/qsimplepopulationsizer.h"
#include "populationsizer/qsimplepopulationsizerfibonacci.h"

#include "crossover/qintegercrossover.h"
#include "crossover/qstringcrossover.h"
#include "crossover/qrccrossover.h"

#include "mutation/qintegermutation.h"
#include "mutation/qstringmutation.h"
#include "mutation/qrcmutation.h"

#include <iostream>
#include <process.h>
#include <vector>

using namespace ECWorkshop;

///////////////////////////////////////////////////////////
//
//		QEAFactory class
//
///////////////////////////////////////////////////////////

//---------------------------------------------------------
QEAFactory::QEAFactory (void) {
}

//---------------------------------------------------------
QEAFactory::QEAFactory (const QEvolutionaryAlgorithm* argOwner)
: _owner(argOwner) {
}

//---------------------------------------------------------
QEAFactory::QEAFactory (const QEAFactory& argFactory) {
}

//---------------------------------------------------------
QEAFactory::~QEAFactory (void) {
	deletePopulations ();
	deleteInitializers ();
	deleteSelections ();
	deleteParentSelectors ();
	deletePopulationSizers ();
	deleteCrossovers ();
	deleteMutations ();
}

//---------------------------------------------------------
QEAFactory& QEAFactory::operator= (const QEAFactory& argFactory) {
	return (*this);
}

//---------------------------------------------------------
QPopulation* QEAFactory::createPopulation (const QEncodingType& argEncoding) {
	QPopulation* tempPopul = NULL;

	switch (argEncoding) {
		case QET_INTEGER:
			tempPopul = new QIntegerPopulation;
			break;
		case QET_STRING:
			tempPopul = new QStringPopulation;
			break;
		case QET_REAL:
			tempPopul = new QRCPopulation;
			break;
		case QET_REGULATORY_NETWORK:
			tempPopul = new QRNPopulation;
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createPopulation error:\n"
				<<"\tA trial to create QPopulation object for undefined (or unspecified) type of encoding\n\n";
			exit (0);
			break;
	}

	_populations.push_back (tempPopul);

	return tempPopul;
}

//---------------------------------------------------------
QInitializer* QEAFactory::createInitializer (const QEncodingType& argEncoding) {
	QInitializer* tempInit = NULL;

	switch (argEncoding) {
		case QET_INTEGER:
			tempInit = new QIntegerInitializer (_owner);
			break;
		case QET_STRING:
			tempInit = new QStringInitializer (_owner);
			break;
		case QET_REAL:
			tempInit = new QRCInitializer (_owner);
			break;
		case QET_REGULATORY_NETWORK:
			tempInit = new QRNInitializer (_owner);
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createInitializer error:\n"
				<<"\tA trial to create QInitializer object for undefined (or unspecified) type of encoding\n\n";
			exit (0);
			break;
	}

	_inits.push_back (tempInit);

	return tempInit;
}

//---------------------------------------------------------
QNichingOperator* QEAFactory::createNichingOperator (const QNichingType& argNichingType) {
	QNichingOperator* tempNiching = NULL;

	switch (argNichingType) {
		case QNT_SHARING:
			tempNiching = new QFitnessSharing (_owner);
			break;
		case QNT_CROWDING_DETERMINISTIC:
			tempNiching = new QDeterministicCrowding (_owner);
			break;
		default:	// QET_UNKNOWN
			//cerr<<"\n\nQEAFactory::createNichingOperator error:\n"
			//	<<"\tA trial to create undefined (or unspecified) type of QNichingOperator object\n\n";
			//exit (0);
			break;
	}

	_nichings.push_back (tempNiching);

	return tempNiching;
}

//---------------------------------------------------------
QSelection* QEAFactory::createSelection (const QSelectionType& argSelType) {
	QSelection* tempSel = NULL;

	switch (argSelType) {
		case QST_NEUTRAL:
			tempSel = new QNeutralSelection (_owner);
			break;
		case QST_TOURNAMENT:
			tempSel = new QTournamentSelection (_owner);
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createSelection error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QSelection object\n\n";
			exit (0);
			break;
	}

	_selections.push_back (tempSel);

	return tempSel;
}

//---------------------------------------------------------
QParentSelector* QEAFactory::createParentSelector (const QParentSelectorType& argPSelType) {
	QParentSelector* tempPSel = NULL;

	switch (argPSelType) {
		case QPST_RANDOM:
			tempPSel = new QRandomParentSelector (_owner);
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createParentSelector error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QParentSelector object\n\n";
			exit (0);
			break;
	}

	_pSelectors.push_back (tempPSel);

	return tempPSel;
}

//---------------------------------------------------------
QPopulationSizer* QEAFactory::createPopulationSizer (const QPopulationSizerType& argSizerType) {
	QPopulationSizer* tempSizer = NULL;

	switch (argSizerType) {
		case QPSZT_CONST:
			tempSizer = new QConstPopulationSizer (_owner);
			break;
		case QPSZT_SIMPLE:
			tempSizer = new QSimplePopulationSizer (_owner);
			break;
		case QPSZT_SIMPLE_FIBONACCI:
			tempSizer = new QSimplePopulationSizerFibonacci (_owner);
			break;
		case QPSZT_STEADY:
			tempSizer = new QSteadyPopulationSizer (_owner);
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createPopulationSizer error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QPopulationSizer object\n\n";
			exit (0);
			break;
	}

	_sizers.push_back (tempSizer);

	return tempSizer;
}

//---------------------------------------------------------
QCrossoverOperator* QEAFactory::createCrossover (const QEncodingType& argEncoding, const QCrossoverType& argXType) {
	QCrossoverOperator* tempX = NULL;

	switch (argEncoding) {
		case QET_INTEGER:
			tempX = createIntegerCrossover (argXType);
			break;
		case QET_STRING:
			tempX = createStringCrossover (argXType);
			break;
		case QET_REAL:
			tempX = createRCCrossover (argXType);
			break;
		case QET_REGULATORY_NETWORK:
			// ignore
//			tempX = createRNCrossover (argXType);
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createCrossover error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QCrossoverOperator object\n\n";
			exit (0);
			break;
	}

	_crossovers.push_back (tempX);

	return tempX;
}

//---------------------------------------------------------
QMutationOperator* QEAFactory::createMutation (const QEncodingType& argEncoding, const QMutationType& argMType) {
	QMutationOperator* tempMut = NULL;

	switch (argEncoding) {
		case QET_INTEGER:
			tempMut = createIntegerMutation (argMType);
			break;
		case QET_STRING:
			tempMut = createStringMutation (argMType);
			break;
		case QET_REAL:
			tempMut = createRCMutation (argMType);
			break;
		case QET_REGULATORY_NETWORK:
			tempMut = createRNMutation (argMType);
			break;
		default:	// QET_UNKNOWN
			cerr<<"\n\nQEAFactory::createMutation error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QMutationOperator object\n\n";
			exit (0);
			break;
	}

	_mutations.push_back (tempMut);

	return tempMut;
}

//---------------------------------------------------------
int QEAFactory::deletePopulations () {

	__VECTOR_POPULATION_PTR::iterator iter = _populations.begin();

	for (; iter != _populations.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_populations.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteInitializers () {

	__VECTOR_INITIALIZER_PTR::iterator iter = _inits.begin();

	for (; iter != _inits.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_inits.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteNichings () {

	__VECTOR_NICHING_PTR::iterator iter = _nichings.begin();

	for (; iter != _nichings.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_nichings.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteSelections () {

	__VECTOR_SELECTION_PTR::iterator iter = _selections.begin();

	for (; iter != _selections.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_selections.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteParentSelectors () {

	__VECTOR_PARENT_SELECTOR_PTR::iterator iter = _pSelectors.begin();

	for (; iter != _pSelectors.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_pSelectors.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deletePopulationSizers () {

	__VECTOR_POPULATION_SIZER_PTR::iterator iter = _sizers.begin();

	for (; iter != _sizers.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_sizers.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteCrossovers () {

	__VECTOR_CROSSOVER_PTR::iterator iter = _crossovers.begin();

	for (; iter != _crossovers.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_crossovers.clear ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteMutations () {

	__VECTOR_MUTATION_PTR::iterator iter = _mutations.begin();

	for (; iter != _mutations.end(); iter++) {
		if ((*iter) != NULL) {
			delete (*iter);
			(*iter) = NULL;
		}
	}
	_mutations.clear ();

	return Q_OK;
}

//---------------------------------------------------------
QCrossoverOperator* QEAFactory::createIntegerCrossover (const QCrossoverType& argXType) {
	QCrossoverOperator* tempX = NULL;

	switch (argXType) {
		case QXT_1POINT:
			tempX = new QIntegerCrossover1 (_owner);
			break;
		case QXT_2POINT:
			tempX = new QIntegerCrossover2 (_owner);
			break;
		case QXT_UNIFORM:
			tempX = new QIntegerCrossover2 (_owner);
			break;
		case QXT_1POINT_PGX:
			tempX = new QIntegerCrossoverPGX1 (_owner);
			break;
		case QXT_2POINT_PGX:
			tempX = new QIntegerCrossoverPGX2 (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createIntegerCrossover error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QCrossoverOperator object (argXType = "<<argXType<<")\n\n";
			exit (0);
	}

	return tempX;
}

//---------------------------------------------------------
QMutationOperator* QEAFactory::createIntegerMutation (const QMutationType& argMType) {
	QMutationOperator* tempMut = NULL;

	switch (argMType) {
		case QMT_BITFLIP:
			tempMut = new QIntegerMutationBitFlip (_owner);
			break;
		case QMT_GAUSS:
			tempMut = new QRCMutationGauss (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createIntegerMutation error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QMutationOperator object (argMType = "<<argMType<<")\n\n";
			exit (0);
	}

	return tempMut;
}

//---------------------------------------------------------
QCrossoverOperator* QEAFactory::createStringCrossover (const QCrossoverType& argXType) {
	QCrossoverOperator* tempX = NULL;

	switch (argXType) {
		case QXT_1POINT:
			tempX = new QStringCrossover1 (_owner);
			break;
		case QXT_2POINT:
			tempX = new QStringCrossover2 (_owner);
			break;
		case QXT_UNIFORM:
			tempX = new QStringCrossoverUniform (_owner);
			break;
		case QXT_ORDER:
			tempX = new QStringCrossoverOrder (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createStringCrossover error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QCrossoverOperator object (argXType = "<<argXType<<")\n\n";
			exit (0);
	}

	return tempX;
}

//---------------------------------------------------------
QMutationOperator* QEAFactory::createStringMutation (const QMutationType& argMType) {
	QMutationOperator* tempMut = NULL;

	switch (argMType) {
		case QMT_RANDOMFLIP:
			tempMut = new QStringMutationRandomFlip (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createStringMutation error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QMutationOperator object (argMType = "<<argMType<<")\n\n";
			exit (0);
	}

	return tempMut;
}

//---------------------------------------------------------
QCrossoverOperator* QEAFactory::createRCCrossover (const QCrossoverType& argXType) {
	QCrossoverOperator* tempX = NULL;

	switch (argXType) {
		case QXT_ARITHMETIC:
			tempX = new QRCCrossoverArithmetic (_owner);
			break;
		case QXT_BLX:
			tempX = new QRCCrossoverBLX (_owner);
			break;
		case QXT_SBX:
			tempX = new QRCCrossoverSBX (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createRCCrossover error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QCrossoverOperator object (argXType = "<<argXType<<")\n\n";
			exit (0);
	}

	return tempX;
}

//---------------------------------------------------------
QMutationOperator* QEAFactory::createRCMutation (const QMutationType& argMType) {
	QMutationOperator *tempMut = NULL;

	switch (argMType) {
		case QMT_GAUSS:
			tempMut = new QRCMutationGauss (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createRCMutation error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QMutationOperator object (argMType = "<<argMType<<")\n\n";
			exit (0);
	}

	return tempMut;
}

//---------------------------------------------------------
QMutationOperator* QEAFactory::createRNMutation (const QMutationType& argMType) {
	QMutationOperator *tempMut = NULL;

	switch (argMType) {
		case QMT_RN_BASIC:
			tempMut = new QRNMutationBasic (_owner);
			break;
		default:
			cerr<<"\n\nQEAFactory::createRNMutation error:\n"
				<<"\tA trial to create undefined (or unspecified) type of QMutationOperator object (argMType = "<<argMType<<")\n\n";
			exit (0);
	}

	return tempMut;
}

//---------------------------------------------------------
int QEAFactory::deleteComponents () {

	deletePopulations ();
	deleteInitializers ();
	deleteNichings ();
	deleteSelections ();
	deleteParentSelectors ();
	deletePopulationSizers ();
	deleteCrossovers ();
	deleteMutations ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteLastCreatedPopulation () {
	delete (*(_populations.end()-1));
	_populations.pop_back ();

	return Q_OK;
}

//---------------------------------------------------------
int QEAFactory::deleteLastCreatedMutation () {
	delete (*(_mutations.end()-1));
	_mutations.pop_back ();

	return Q_OK;
}
